/*
 * InSyTrack - software for tracking program flow.
 *	-- server main file.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#include <stdio.h>

/* time headers for select() (the old way) */
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  ifdef HAVE_TIME_H
#   include <time.h>
#  endif
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>	/* select() (the old way), sleep() */
#endif

/* select() - the new way */
#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#if (defined HAVE_GETOPT_H) && (defined HAVE_GETOPT_LONG)
# define _GNU_SOURCE	1	/* getopt_long() */
# include <getopt.h>
#endif

#ifdef HAVE_LIBINTL_H
# include <libintl.h>	/* translation stuff */
#endif

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#ifdef HAVE_GETTEXT
# ifndef _
#  define 	_(String)		gettext (String)
# endif
#else
# define 	_(String)		String
#endif

#define	gettext_noop(String)	String
#define	N_(String)		String

#ifdef HAVE_STRING_H
# if ((!defined STDC_HEADERS) || (!STDC_HEADERS)) && (defined HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_LIBGEN_H
# include <libgen.h>	/* basename() */
#endif

#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>	/* fstat() to check stdout & stderr */
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>	/* htonl(), htons(),  */
#endif

#ifdef HAVE_ARPA_INET_H
# include <arpa/inet.h>		/* inet_pton() */
#endif

#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif

#include "insytrack_server.h"
#include "insytrackd_signal.h"
#include "insytrackd_msg.h"
/* INSYTRACK_HAVE_SELECT, INSYTRACK_MAX_SELECT_SECONDS: */

/**
 * The number of seconds per each select() call.
 */
#define INSYTRACK_MAX_SELECT_SECONDS 1


#if ((defined HAVE_SYS_SELECT_H) || (((defined TIME_WITH_SYS_TIME)	\
	|| (defined HAVE_SYS_TIME_H) || (defined HAVE_TIME_H))		\
	&& (defined HAVE_UNISTD_H)))					\
	&& (defined HAVE_SELECT)
# define INSYTRACK_HAVE_SELECT 1
#else
# undef INSYTRACK_HAVE_SELECT
#endif

#define	INSYTRACK_PROGRAM_NAME	"InSyTrackd" /*PACKAGE*/


static const char insy_ver_str[] = N_("version");
static const char insy_author_str[] = "Copyright (C) 2012-2013 Bogdan 'bogdro' Drozdowski, bogdandr@op.pl\n";
static const char insy_lic_str[] = N_(							\
	"Program flow tracking server.\n"			\
	"\nThis program is Free Software; you can redistribute it and/or"		\
	"\nmodify it under the terms of the GNU Lesser General Public License"			\
	"\nas published by the Free Software Foundation; either version 3"		\
	"\nof the License, or (at your option) any later version."			\
	"\n\nThis program is distributed in the hope that it will be useful,"		\
	"\nbut WITHOUT ANY WARRANTY; without even the implied warranty of"		\
	"\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n");

/* Error messages explaining the stage during which an error occurred. */
const char * const insy_err_msg                   = N_("error");
const char * const err_msg_thread_init            = N_("during thread initialization");
const char * const err_msg_server_init            = N_("during server initialization");
const char * const err_msg_server_start           = N_("during server start");

/* Messages displayed when verbose mode is on */
static const char * const insy_msg_signal         = N_("Setting signal handlers");
static const char * const insy_msg_listen         = N_("Initiating listening");
static const char * const insy_msg_nobg           = N_("Going into background not supported or failed");
static const char * const insy_msg_client_conn    = N_("Client connected");
static const char * const insy_msg_client_disconn = N_("Client disconnected");

/* Command-line options. */
static int insy_opt_bg         = 0;
static int insy_opt_verbose    = 0;

#if (defined HAVE_GETOPT_H) && (defined HAVE_GETOPT_LONG)
static int insy_opt_char       = 0;

static int insy_opt_conn       = 0;
static int insy_opt_help       = 0;
static int insy_opt_license    = 0;
# if (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON)
static int insy_opt_listenaddr = 0;
# endif
# ifdef HAVE_NETINET_IN_H
static int insy_opt_port       = 0;
# endif
static int insy_opt_proto      = 0;
static int insy_opt_socktype   = 0;
static int insy_opt_sockdom    = 0;
static int insy_opt_version    = 0;
/* have to use a temp variable, to add both '-v' and '--verbose' together. */
static int insy_opt_verbose_temp = 0;

static const struct option insy_opts[] =
{
	{ "background",		no_argument,		&insy_opt_bg,		1 },
	{ "connections",	required_argument,	&insy_opt_conn,		1 },
	{ "help",		no_argument,		&insy_opt_help,		1 },
	{ "licence",		no_argument,		&insy_opt_license,	1 },
	{ "license",		no_argument,		&insy_opt_license,	1 },
# if (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON)
	{ "listen",		required_argument,	&insy_opt_listenaddr,	1 },
# endif
# ifdef HAVE_NETINET_IN_H
	{ "port",		required_argument,	&insy_opt_port,		1 },
# endif
	{ "protocol",		required_argument,	&insy_opt_proto,	1 },
	{ "socket-type",	required_argument,	&insy_opt_socktype,	1 },
	{ "socket-domain",	required_argument,	&insy_opt_sockdom,	1 },
	/* have to use a temp variable, to add both '-v' and '--verbose' together. */
	{ "verbose",		no_argument,		&insy_opt_verbose_temp,	1 },
	{ "version",		no_argument,		&insy_opt_version,	1 },
	{ NULL, 0, NULL, 0 }
};
#endif

static const char *insy_progname;	/* The name of the program */
static int insy_stdout_open = 1, insy_stderr_open = 1;

static int insy_socket_type = INSYTRACK_SERVER_SOCKET_TYPE;
static int insy_socket_domain = INSYTRACK_SERVER_SOCKET_DOMAIN;
static int insy_socket_protocol = INSYTRACK_SERVER_SOCKET_PROTOCOL;
static int insy_socket_backlog = INSYTRACK_SERVER_SOCKET_BACKLOG;

#ifdef HAVE_NETINET_IN_H
static struct sockaddr_in insy_default_listen_addr;/* =
{
	AF_INET,
	INSYTRACK_SERVER_SOCKET_PORT,
	0
};*/

#endif
static struct sockaddr * insy_listen_addr = NULL;
static socklen_t insy_addrlen = 0;

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void insytrackd_check_stds INSYTRACK_PARAMS ((int * const stdout_open, int * const stderr_open));
#endif

/**
 * Checks if stdout & stderr are open.
 * \param stdout_open Pointer to an int, which will get the value 0 if standard output is not open.
 * \param stderr_open Pointer to an int, which will get the value 0 if standard error output is not open.
 */
static void
#ifdef INSYTRACK_ANSIC
INSYTRACK_ATTR ((nonnull))
#endif
insytrackd_check_stds (
#ifdef INSYTRACK_ANSIC
	int * const stdout_open, int * const stderr_open)
#else
	stdout_open, stderr_open)
	int * const stdout_open;
	int * const stderr_open;
#endif
{
#ifdef HAVE_SYS_STAT_H
	int res;
	struct stat stat_buf;
#endif

	if ( stdout_open != NULL )
	{
		*stdout_open = 1;

#ifdef HAVE_SYS_STAT_H

# ifdef HAVE_UNISTD_H
		res = fstat (STDOUT_FILENO, &stat_buf);
# else
		res = fstat (1, &stat_buf);
# endif
		if ( res < 0 )
		{
			*stdout_open = 0;
		}
#endif	/* HAVE_SYS_STAT_H */
	}

	if ( stderr_open != NULL )
	{
		*stderr_open = 1;

#ifdef HAVE_SYS_STAT_H
# ifdef HAVE_UNISTD_H
		res = fstat (STDERR_FILENO, &stat_buf);
# else
		res = fstat (2, &stat_buf);
# endif
		if ( res < 0 )
		{
			*stderr_open = 0;
		}
#endif	/* HAVE_SYS_STAT_H */
	}

	if ( (stdout == NULL) && (stdout_open != NULL) )
	{
		*stdout_open = 0;
	}
	if ( (stderr == NULL) && (stderr_open != NULL) )
	{
		*stderr_open = 0;
	}
}

/* ======================================================================== */

static void show_error INSYTRACK_PARAMS ((
	const char * const	msg,
	const char * const	extra));

/**
 * Displays an error message.
 * \param err Error code.
 * \param msg The message.
 * \param extra Last element of the error message.
 */
void
#ifdef INSYTRACK_ANSIC
INSYTRACK_ATTR ((nonnull))
#endif
show_error (
#ifdef INSYTRACK_ANSIC
	const char * const	msg,
	const char * const	extra )
#else
	msg, extra )
	const char * const	msg;
	const char * const	extra;
#endif
{
	if ( (insy_stderr_open == 0) || (msg == NULL) )
	{
		return;
	}

	fprintf ( stderr, "%s: %s %s %s\n", insy_progname,
		_(insy_err_msg),
		_(msg),
		(extra != NULL)? extra : "");
	fflush (stderr);
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void show_msg INSYTRACK_PARAMS ((
	const int		type,
	const char * const	msg,
	const char * const	extra));
#endif

/**
 * Displays a message (verbose mode).
 * \param type Type of message (0 == 2-parameter, 1 == 3-parameter)
 * \param msg The message.
 * \param extra Last element of the error message.
 */
static void
#ifdef INSYTRACK_ANSIC
INSYTRACK_ATTR ((nonnull))
#endif
show_msg (
#ifdef INSYTRACK_ANSIC
	const int		type,
	const char * const	msg,
	const char * const	extra)
#else
	type, msg, extra )
	const int		type;
	const char * const	msg;
	const char * const	extra;
#endif
{
	if ( (insy_stdout_open == 0) || (msg == NULL) )
	{
		return;
	}

	if ( (type == 0) || (extra == NULL) )
	{
		printf ( "%s: %s\n", insy_progname, _(msg) );
	}
	else
	{
		printf ( "%s: %s: '%s'\n", insy_progname, _(msg), extra );
	}
	fflush (stdout);
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void print_help INSYTRACK_PARAMS ((const char * const my_name));
#endif

/**
 * Prints the help screen.
 * \param my_name Program identifier, like argv[0], if available.
 */
static void
#ifdef INSYTRACK_ANSIC
INSYTRACK_ATTR ((nonnull))
#endif
print_help (
#ifdef INSYTRACK_ANSIC
	const char * const my_name )
#else
	my_name )
	const char * const my_name;
#endif
{
	const char *prog;
	if ( my_name == NULL )
	{
		prog = INSYTRACK_PROGRAM_NAME;
#ifdef HAVE_STRING_H
	}
	else if ( my_name[0] == '\0' /*strlen (my_name) == 0*/ )
	{
		prog = INSYTRACK_PROGRAM_NAME;
#endif
	}
	else
	{
		prog = my_name;
	}

	printf ("%s", prog);
	printf ( "%s", _(" - Program flow tracking server\nSyntax: ") );
	printf ("%s", prog);
	printf ( "%s\n", _(" [options] ") );
	puts ( _("Options:") );
	puts ( _("--background\t\tContinue work in the background, if possible") );
	puts ( _("--connections <n>\tAllow up to <n> connections waiting to be serviced") );
	puts ( _("-h|--help\t\tPrint help") );
	puts ( _("-l|--license\t\tPrint license information") );
#if (defined HAVE_NETINET_IN_H) && (defined HAVE_INET_PTON)
	puts ( _("--listen <addr>\t\tListen on the given address only") );
#endif
#ifdef HAVE_NETINET_IN_H
	puts ( _("--port <port>\t\tListen on the given port") );
#endif
	puts ( _("--protocol <n>\t\tUse protocol number <n> for the socket") );
	puts ( _("--socket-domain <n>\tUse socket domain <n> for the socket") );
	puts ( _("--socket-type <n>\tUse socket type <n> for the socket") );
	puts ( _("-v|--verbose\t\tVerbose output") );
	puts ( _("-V|--version\t\tPrint version number") );
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void insy_put_addr INSYTRACK_PARAMS ((
	const struct sockaddr * const insy_addr,
	char * const insy_ret_buf,
	const size_t insy_ret_buf_len));
#endif

static void insy_put_addr (
#ifdef INSYTRACK_ANSIC
	const struct sockaddr * const insy_addr,
	char * const insy_ret_buf,
	const size_t insy_ret_buf_len)
#else
	insy_addr, insy_ret_buf, insy_ret_buf_len)
	const struct sockaddr * const insy_addr;
	char * const insy_ret_buf;
	const size_t insy_ret_buf_len;
#endif
{
#if (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON)
# if ((defined HAVE_SNPRINTF) || (defined HAVE_SPRINTF))
	size_t insy_cur_len;
# endif
	const struct sockaddr_in * insy_inet_addr;
	const struct sockaddr_in6 * insy_inet_addr6;

	if ( (insy_addr == NULL) || (insy_ret_buf == NULL)
		|| (insy_ret_buf_len == 0) )
	{
		return;
	}
	insy_inet_addr = (const struct sockaddr_in *)insy_addr;
	insy_inet_addr6 = (const struct sockaddr_in6 *)insy_addr;

	insy_ret_buf[0] = '\0';
	if ( insy_addr->sa_family == AF_INET )
	{
		inet_ntop (insy_inet_addr->sin_family,
			&(insy_inet_addr->sin_addr.s_addr), insy_ret_buf,
			insy_ret_buf_len-1);
		insy_ret_buf[insy_ret_buf_len-1] = '\0';
# ifdef HAVE_SNPRINTF
		insy_cur_len = strlen (insy_ret_buf);
		snprintf (&insy_ret_buf[insy_cur_len],
			insy_ret_buf_len - 1 - insy_cur_len,
			":%d", htons (insy_inet_addr->sin_port)); /* warnings inside htons() */
# else
#  ifdef HAVE_SPRINTF
		insy_cur_len = strlen (insy_ret_buf);
		sprintf (&insy_ret_buf[insy_cur_len],
			":%d", htons (insy_inet_addr->sin_port)); /* warnings inside htons() */
#  endif
# endif
	}
	else if ( insy_addr->sa_family == AF_INET6 )
	{
		inet_ntop (insy_inet_addr6->sin6_family,
			&(insy_inet_addr6->sin6_addr.s6_addr), insy_ret_buf,
			insy_ret_buf_len-1);
		insy_ret_buf[insy_ret_buf_len-1] = '\0';
# ifdef HAVE_SNPRINTF
		insy_cur_len = strlen (insy_ret_buf);
		snprintf (&insy_ret_buf[insy_cur_len],
			insy_ret_buf_len - 1 - insy_cur_len,
			":%d", htons (insy_inet_addr6->sin6_port)); /* warnings inside htons() */
# else
#  ifdef HAVE_SPRINTF
		insy_cur_len = strlen (insy_ret_buf);
		sprintf (&insy_ret_buf[insy_cur_len],
			":%d", htons (insy_inet_addr6->sin6_port)); /* warnings inside htons() */
#  endif
# endif
	}
#endif
	insy_ret_buf[insy_ret_buf_len-1] = '\0';
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void client_connected INSYTRACK_PARAMS ((
	const struct sockaddr * const insy_addr,
	const socklen_t insy_addr_len, void * insy_param));
#endif

static void client_connected (
#ifdef INSYTRACK_ANSIC
	const struct sockaddr * const insy_addr,
	const socklen_t insy_addr_len INSYTRACK_ATTR ((unused)),
	void * insy_param INSYTRACK_ATTR ((unused)))
#else
	insy_addr, insy_addr_len, insy_param)
	const struct sockaddr * const insy_addr;
	const socklen_t insy_addr_len  INSYTRACK_ATTR ((unused));
	void * insy_param INSYTRACK_ATTR ((unused));
#endif
{
	char insy_conn_addr[200];

	if ( insy_opt_verbose > 0 )
	{
#ifdef HAVE_PTHREAD_H
		pthread_mutex_lock (&insy_display_mutex);
#endif
		insy_put_addr (insy_addr, insy_conn_addr,
			sizeof (insy_conn_addr));
		show_msg ( 1, insy_msg_client_conn, insy_conn_addr );
#ifdef HAVE_PTHREAD_H
		pthread_mutex_unlock (&insy_display_mutex);
#endif
	}
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void client_disconnected INSYTRACK_PARAMS ((
	const struct sockaddr * const insy_addr,
	const socklen_t insy_addr_len, void * insy_param));
#endif

static void client_disconnected (
#ifdef INSYTRACK_ANSIC
	const struct sockaddr * const insy_addr,
	const socklen_t insy_addr_len INSYTRACK_ATTR ((unused)),
	void * insy_param INSYTRACK_ATTR ((unused)))
#else
	insy_addr, insy_addr_len, insy_param)
	const struct sockaddr * const insy_addr;
	const socklen_t insy_addr_len INSYTRACK_ATTR ((unused));
	void * insy_param INSYTRACK_ATTR ((unused));
#endif
{
	char insy_disconn_addr[200];

	if ( insy_opt_verbose > 0 )
	{
#ifdef HAVE_PTHREAD_H
		pthread_mutex_lock (&insy_display_mutex);
#endif
		insy_put_addr (insy_addr, insy_disconn_addr,
			sizeof (insy_disconn_addr));
		show_msg ( 1, insy_msg_client_disconn, insy_disconn_addr );
#ifdef HAVE_PTHREAD_H
		pthread_mutex_unlock (&insy_display_mutex);
#endif
	}
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
int main INSYTRACK_PARAMS ((int argc, char* argv[]));
#endif

/**
 * The entry point of the InSyTrack server.
 * @return an error code for the calling process.
 */
int main (
#ifdef INSYTRACK_ANSIC
	int argc, char* argv[] )
#else
	argc, argv )
	int argc;
	char* argv[];
#endif
{
	insytrack_server_t insy_server;
	insytrack_server_error_t insy_server_error;
#if (!defined HAVE_SLEEP) && (defined INSYTRACK_HAVE_SELECT)
	struct timeval insy_tv;
	fd_set insy_set;
#endif
#if ! ((defined HAVE_GETOPT_H) && (defined HAVE_GETOPT_LONG))
	int i;
#endif
	int insy_optind = 0;
	int res;
	int port_res;
	char listen_addr[200];

	/* internationalization: */
#ifdef HAVE_LIBINTL_H
# ifdef HAVE_SETLOCALE
	setlocale (LC_ALL, "");
# endif
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);
#endif
	insytrackd_check_stds (&insy_stdout_open, &insy_stderr_open);

	insy_progname = NULL;
	if ( argv[0] != NULL )
	{
#if (defined HAVE_LIBGEN_H) && (defined HAVE_BASENAME)
		insy_progname = basename (argv[0]);
#else
# if (defined HAVE_STRING_H)
		insy_progname = strrchr (argv[0], (int)'/') + 1;
# else
		insy_progname = argv[0];
# endif
#endif
		if ( insy_progname == NULL )
		{
			insy_progname = INSYTRACK_PROGRAM_NAME;
		}
	}
	else
	{
		insy_progname = INSYTRACK_PROGRAM_NAME;
	}

	/* initialization with default values: */
#ifdef HAVE_NETINET_IN_H
	insy_default_listen_addr.sin_family = AF_INET;
	insy_default_listen_addr.sin_port = htons (INSYTRACK_SERVER_SOCKET_PORT); /* warnings inside */
	insy_default_listen_addr.sin_addr.s_addr = htonl (INADDR_ANY); /* warnings inside */
	insy_listen_addr = (struct sockaddr *) &insy_default_listen_addr;
	insy_addrlen = sizeof (struct sockaddr_in);
#endif

	/* command line: */
#if (defined HAVE_GETOPT_H) && (defined HAVE_GETOPT_LONG)
	optind = 0;
	while (1==1)
	{
		insy_opt_char = getopt_long ( argc, argv, "Vhlv", insy_opts, NULL );
		if ( insy_opt_char == -1 )
		{
			break;
		}

		/* NOTE: these shouldn't be a sequence of else-ifs */
		if ( (insy_opt_char == (int)'?') || (insy_opt_char == (int)':') )
		{
			if ( insy_stdout_open == 1 )
			{
				print_help (insy_progname);
			}
			return INSYTRACK_EXIT_OK;
		}

		if ( (insy_opt_char == (int)'h') || (insy_opt_help == 1) )
		{
			if ( insy_stdout_open == 1 )
			{
				print_help (insy_progname);
			}
			return INSYTRACK_EXIT_OK;
		}

		if ( (insy_opt_char == (int)'V') || (insy_opt_version == 1) )
		{
			show_msg ( 1, insy_ver_str, VERSION );
			return INSYTRACK_EXIT_OK;
		}

		if ( (insy_opt_char == (int)'l') || (insy_opt_license == 1) )
		{
			if ( insy_stdout_open == 1 )
			{
				show_msg ( 0, insy_lic_str, "" );
				puts ( insy_author_str );
			}
			return INSYTRACK_EXIT_OK;
		}

		if ( (insy_opt_char == (int)'v') /* do NOT check for opt_verbose here */ )
		{
			insy_opt_verbose++;
		}

		if ( insy_opt_conn == 1 )
		{
			res = sscanf ( optarg, "%d", &insy_socket_backlog );
			if ( (res == 0) || (insy_socket_backlog < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			insy_opt_conn = 0;
		}

		if ( insy_opt_proto == 1 )
		{
			res = sscanf ( optarg, "%d", &insy_socket_protocol );
			if ( (res == 0) || (insy_socket_protocol < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			insy_opt_proto = 0;
		}

		if ( insy_opt_socktype == 1 )
		{
			res = sscanf ( optarg, "%d", &insy_socket_type );
			if ( (res == 0) || (insy_socket_type < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			insy_opt_socktype = 0;
		}

		if ( insy_opt_sockdom == 1 )
		{
			res = sscanf ( optarg, "%d", &insy_socket_domain );
			if ( (res == 0) || (insy_socket_domain < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
# ifdef HAVE_NETINET_IN_H
			insy_default_listen_addr.sin_family
				= (sa_family_t)insy_socket_domain;
# endif
			insy_opt_sockdom = 0;
		}
# if (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON)
		if ( insy_opt_listenaddr == 1 )
		{
			res = inet_pton (insy_default_listen_addr.sin_family,
				optarg, &(insy_default_listen_addr.sin_addr.s_addr));
			if ( res != 1 )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			insy_opt_listenaddr = 0;
		}
# endif /* (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON) */
# ifdef HAVE_NETINET_IN_H
		if ( insy_opt_port == 1 )
		{
			res = sscanf ( optarg, "%d", &port_res );
			if ( (res == 0) || (port_res < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			insy_default_listen_addr.sin_port = (in_port_t)htons (port_res); /* warnings inside */
			insy_opt_port = 0;
		}
# endif /* HAVE_NETINET_IN_H */
	}
	insy_optind = optind;
	/* add up '-v' and '--verbose'. */
	insy_opt_verbose += insy_opt_verbose_temp;

#else	/* no getopt_long */

	for ( i = 1; i < argc; i++ )	/* argv[0] is the program name */
	{
		if ( argv[i] == NULL )
		{
			continue;
		}
		/* NOTE: these shouldn't be a sequence of else-ifs */
		if ( (strcmp (argv[i], "-h") == 0) || (strcmp (argv[i], "-?") == 0)
			|| (strcmp (argv[i], "--help") == 0) )
		{
			if ( insy_stdout_open == 1 )
			{
				print_help (insy_progname);
			}
			return INSYTRACK_EXIT_OK;
		}

		if ( (strcmp (argv[i], "-V") == 0) || (strcmp (argv[i], "--version") == 0) )
		{
			show_msg ( 1, insy_ver_str, VERSION );
			return INSYTRACK_EXIT_OK;
		}

		if ( (strcmp (argv[i], "-l") == 0) || (strcmp (argv[i], "--licence") == 0)
			|| (strcmp (argv[i], "--license") == 0) )
		{
			if ( insy_stdout_open == 1 )
			{
				show_msg ( 0, insy_lic_str, "" );
				puts ( insy_author_str );
			}
			return INSYTRACK_EXIT_OK;
		}

		if ( (strcmp (argv[i], "-v") == 0) || (strcmp (argv[i], "--verbose") == 0) )
		{
			insy_opt_verbose++;
			argv[i] = NULL;
			continue;
		}

		if ( strcmp (argv[i], "--background") == 0 )
		{
			insy_opt_bg = 1;
			argv[i] = NULL;
			continue;
		}
		if ( strcmp (argv[i], "--") == 0 )
		{
			/* end-of-arguments marker */
			argv[i] = NULL;
			break;
		}

		if ( strcmp (argv[i], "--connections") == 0 )
		{
			if ( i+1 >= argc )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}

			res = sscanf ( argv[i+1], "%d", &insy_socket_backlog );
			if ( (res == 0) || (insy_socket_backlog < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			argv[i] = NULL;
			argv[i+1] = NULL;
			continue;
		}

		if ( strcmp (argv[i], "--protocol") == 0 )
		{
			if ( i+1 >= argc )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}

			res = sscanf ( argv[i+1], "%d", &insy_socket_protocol );
			if ( (res == 0) || (insy_socket_protocol < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			argv[i] = NULL;
			argv[i+1] = NULL;
			continue;
		}

		if ( strcmp (argv[i], "--socket-type") == 0 )
		{
			if ( i+1 >= argc )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}

			res = sscanf ( argv[i+1], "%d", &insy_socket_type );
			if ( (res == 0) || (insy_socket_type < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			argv[i] = NULL;
			argv[i+1] = NULL;
			continue;
		}

		if ( strcmp (argv[i], "--socket-domain") == 0 )
		{
			if ( i+1 >= argc )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}

			res = sscanf ( argv[i+1], "%d", &insy_socket_domain );
			if ( (res == 0) || (insy_socket_domain < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
# ifdef HAVE_NETINET_IN_H
			insy_default_listen_addr.sin_family
				= (sa_family_t)insy_socket_domain;
# endif
			argv[i] = NULL;
			argv[i+1] = NULL;
			continue;
		}
# if (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON)
		if ( strcmp (argv[i], "--listen") == 0 )
		{
			if ( i+1 >= argc )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}

			res = inet_pton (insy_default_listen_addr.sin_family,
				argv[i+1], &(insy_default_listen_addr.sin_addr.s_addr));
			if ( res != 1 )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			argv[i] = NULL;
			argv[i+1] = NULL;
			continue;
		}
# endif /* (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_PTON) */
# ifdef HAVE_NETINET_IN_H
		if ( strcmp (argv[i], "--port") == 0 )
		{
			if ( i+1 >= argc )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}

			res = sscanf ( argv[i+1], "%d", &port_res );
			if ( (res == 0) || (port_res < 0) )
			{
				if ( insy_stdout_open == 1 )
				{
					print_help (insy_progname);
				}
				return INSYTRACK_EXIT_BAD_CMDLINE;
			}
			insy_default_listen_addr.sin_port = (in_port_t)htons (port_res); /* warnings inside */
			argv[i] = NULL;
			argv[i+1] = NULL;
			continue;
		}
# endif /* HAVE_NETINET_IN_H */
	}
	insy_optind = 1;
#endif /* getopt */

#ifdef __GNUC__
# pragma GCC poison optind
#endif

	if ( insy_stdout_open == 0 )
	{
		insy_opt_verbose = 0;
	}

	/* daemon mode: */
	if ( insy_opt_bg == 1 )
	{
#ifdef HAVE_DAEMON
		if ( daemon (1, 1) != 0 )
		{
			puts ( insy_msg_nobg );
		}
		/* don't disable verbose mode, because the output
		   can be redirected to a file */
		/*else
		{
			insy_opt_verbose = 0;
			insy_stdout_open = 0;
			insy_stderr_open = 0;
		}*/
#else
		puts ( insy_msg_nobg );
#endif
	}

	/* setting signal handlers: */
#ifdef HAVE_SIGNAL_H
	if ( insy_opt_verbose > 0 )
	{
		show_msg ( 0, insy_msg_signal, "" );
	}
	insytrackd_set_sigh ();
#endif		/* HAVE_SIGNAL_H */


	res = pthread_mutex_init (&insy_display_mutex, NULL);
	if ( res != 0 )
	{
		show_error ( err_msg_thread_init, "" );
		return INSYTRACK_EXIT_MUTEX;
	}

	insy_server = insytrack_server_create ();
	if ( insy_server == INSYTRACK_SERVER_INVALID )
	{
		pthread_mutex_destroy (&insy_display_mutex);
		show_error ( err_msg_server_init, "" );
		return INSYTRACK_EXIT_SERVER_ERROR;
	}
	insytrack_server_set_msgrcvd_fcn (insy_server,
		&insytrackd_msg_received, NULL);
	insytrack_server_set_conn_fcn (insy_server,
		&client_connected, NULL);
	insytrack_server_set_disconn_fcn (insy_server,
		&client_disconnected, NULL);
	if ( (insy_listen_addr != NULL) && (insy_addrlen != 0) )
	{
		insytrack_server_set_addr (insy_server, insy_listen_addr, insy_addrlen);
	}
	insytrack_server_set_sockdomain (insy_server, insy_socket_domain);
	insytrack_server_set_socktype (insy_server, insy_socket_type);
	insytrack_server_set_sockproto (insy_server, insy_socket_protocol);
	insytrack_server_set_backlog (insy_server, insy_socket_backlog);

#if (defined HAVE_NETINET_IN_H) && (defined HAVE_ARPA_INET_H) && (defined HAVE_INET_NTOP)
	if ( insy_opt_verbose > 0 )
	{
		insy_put_addr ((struct sockaddr *)&insy_default_listen_addr,
			listen_addr, sizeof (listen_addr));
		show_msg ( 1, insy_msg_listen, listen_addr );
	}
#endif
	insy_server_error = insytrack_server_start (insy_server);
	if ( insy_server_error != INSYTRACKS_OK )
	{
		insytrack_server_close (insy_server);
		show_error ( err_msg_server_start, "" );
		return insy_server_error;
	}

	while (insytrackd_sig_recvd == 0)
	{
#ifdef HAVE_SLEEP
		sleep (INSYTRACK_MAX_SELECT_SECONDS);
#else
# ifdef INSYTRACK_HAVE_SELECT
		/* select() can destroy its parameters, so reset every time */
		FD_ZERO (&insy_set);
		insy_tv.tv_sec = INSYTRACK_MAX_SELECT_SECONDS;
		insy_tv.tv_usec = 0;
		select (0, &insy_set, NULL, NULL, &insy_tv);
# else
		volatile int i;
		for ( i = 0; i < (1<<30); i++ );
# endif
#endif
	}
	insytrack_server_close (insy_server);
	pthread_mutex_destroy (&insy_display_mutex);
	return INSYTRACK_EXIT_OK;
}
