/*
 * InSyTrack - software for tracking program flow.
 *	-- processing received messages.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
# if ((!defined STDC_HEADERS) || (!STDC_HEADERS)) && (defined HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_GETTIMEOFDAY
# if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
# else
#  if HAVE_SYS_TIME_H
#   include <sys/time.h>
#  else
#   ifdef HAVE_TIME_H
#    include <time.h>
#   endif
#  endif
# endif
#endif

#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif

#include "insytrack_server.h"
#include "insytrackd_msg.h"

#ifdef HAVE_PTHREAD_H
/**
 * The mutex for displaying the received messages (so that no two
 * threads can corrupt each other's output).
 */
pthread_mutex_t insy_display_mutex;
#endif

/* ======================================================================== */

/**
 * Processes a received message.
 * @param insy_msg The message to process.
 * @param insy_data User-provided data.
 * @return INSYTRACK_CALLBACK_ABORT or INSYTRACK_CALLBACK_CONTINUE
 */
insytrack_callback_return_t
insytrackd_msg_received (
#ifdef INSYTRACK_ANSIC
	insytrack_message_t * insy_msg,
	void * insy_data INSYTRACK_ATTR ((unused)))
#else
	insy_msg, insy_data)
	insytrack_message_t * insy_msg;
	void * insy_data INSYTRACK_ATTR ((unused));
#endif
{
#if (defined HAVE_GETTIMEOFDAY) && (defined HAVE_LOCALTIME) \
	&& (defined HAVE_STRFTIME)
	struct timeval insy_recv_time;
	struct timezone insy_recv_timezone;
	int insy_timeres;
	struct tm * insy_tm;
	char insy_strtime[200];
# ifdef HAVE_SNPRINTF
	size_t insy_curr_len;
# endif
#endif

	if ( insy_msg == NULL )
	{
		return INSYTRACK_CALLBACK_ABORT;
	}

	if ( insy_msg->insy_msgid_len.insy_msgid_len_int
		< INSYTRACK_MSGID_MAXLEN )
	{
		((char *)insy_msg->insy_msgid)[
			insy_msg->insy_msgid_len.insy_msgid_len_int]
			= '\0';
	}
	else
	{
		((char *)insy_msg->insy_msgid)[
			INSYTRACK_MSGID_MAXLEN] = '\0';
	}
	if ( insy_msg->insy_msgbody_len.insy_msgbody_len_int
		< INSYTRACK_MSGBODY_MAXLEN )
	{
		((char *)insy_msg->insy_msgbody)[
			insy_msg->insy_msgbody_len.insy_msgbody_len_int]
			= '\0';
	}
	else
	{
		((char *)insy_msg->insy_msgbody)[
			INSYTRACK_MSGBODY_MAXLEN]
			= '\0';
	}
#ifdef HAVE_PTHREAD_H
	pthread_mutex_lock (&insy_display_mutex);
#endif

#if (defined HAVE_GETTIMEOFDAY) && (defined HAVE_LOCALTIME) \
	&& (defined HAVE_STRFTIME)
	insy_timeres = gettimeofday (&insy_recv_time, &insy_recv_timezone);
	if ( insy_timeres == 0 )
	{
		insy_tm = localtime (&(insy_recv_time.tv_sec));
		if ( insy_tm != NULL )
		{
			insy_strtime[0] = '\0';
			strftime (insy_strtime, sizeof(insy_strtime) - 1,
				"%Y-%m-%d %H:%M:%S", insy_tm);
			insy_strtime[sizeof(insy_strtime) - 1] = '\0';
# ifdef HAVE_SNPRINTF
			insy_curr_len = strlen (insy_strtime);
			snprintf (&insy_strtime[insy_curr_len],
				sizeof(insy_strtime) - 1 - insy_curr_len,
				".%ld", insy_recv_time.tv_usec);
# endif
			insy_strtime[sizeof(insy_strtime) - 1] = '\0';
			printf("%s: ", insy_strtime);
		}
	}
#endif
	printf ("Got message of type %d with id='%s'" \
		" (length %d) and body='%s' (length %d)\n",
		insy_msg->insy_type.insy_type_int,
		(insy_msg->insy_msgid != NULL)? (char *)insy_msg->insy_msgid : "",
		insy_msg->insy_msgid_len.insy_msgid_len_int,
		(insy_msg->insy_msgbody != NULL)? (char *)insy_msg->insy_msgbody : "",
		insy_msg->insy_msgbody_len.insy_msgbody_len_int);

#ifdef HAVE_PTHREAD_H
	pthread_mutex_unlock (&insy_display_mutex);
#endif

	return INSYTRACK_CALLBACK_CONTINUE;
}
