/*
 * InSyTrack - software for tracking program flow, client files.
 *	-- client initialization & shutdown.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>	/* htonl(), htons() */
#endif

#ifdef HAVE_ARPA_INET_H
# include <arpa/inet.h>		/* htonl(), htons() */
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_MALLOC_H
# include <malloc.h>
#endif

#include "insytrack_client.h"
#include "insytrack_server.h"
#include "libinsytrack_client_init.h"

/* ====================================================================== */

/**
 * Creates a new client.
 * @return a client handle with the default configuration
 *	or INSYTRACK_CLIENT_INVALID in case of error.
 */
insytrack_client_t
insytrack_client_create (
#ifdef INSYTRACK_ANSIC
	void
#endif
)
{
	int insy_result;
	struct insytrack_client * insy_ret;

	insy_ret = (struct insytrack_client *) malloc (sizeof (struct insytrack_client));
	if ( insy_ret == NULL )
	{
		return INSYTRACK_CLIENT_INVALID;
	}
	insy_result = pthread_mutex_init (&(insy_ret->insy_client_mutex), NULL);
	if ( insy_result != 0 )
	{
		free (insy_ret);
		return INSYTRACK_CLIENT_INVALID;
	}
	insy_ret->insy_client_socket = -1;
	insy_ret->insy_socket_domain = INSYTRACK_CLIENT_SOCKET_DOMAIN;
	insy_ret->insy_socket_type = INSYTRACK_CLIENT_SOCKET_TYPE;
	insy_ret->insy_socket_protocol = INSYTRACK_CLIENT_SOCKET_PROTOCOL;
#ifdef HAVE_NETINET_IN_H
	insy_ret->insy_socket_connaddr =
		(struct sockaddr *) malloc (sizeof (struct sockaddr_in));
	if ( insy_ret->insy_socket_connaddr != NULL )
	{
		((struct sockaddr_in *)(insy_ret->insy_socket_connaddr))->sin_family
			= AF_INET;
		((struct sockaddr_in *)(insy_ret->insy_socket_connaddr))->sin_port
			= htons (INSYTRACK_SERVER_SOCKET_PORT); /* warnings inside */
		((struct sockaddr_in *)(insy_ret->insy_socket_connaddr))->sin_addr.s_addr
			= htonl (INADDR_LOOPBACK); /* warnings inside */
		insy_ret->insy_connaddr_len = sizeof (struct sockaddr_in);
		insy_ret->insy_connaddr_alloc = 1;
	}
	else
	{
		insy_ret->insy_connaddr_len = 0;
		insy_ret->insy_connaddr_alloc = 0;
	}
#else
	insy_ret->insy_socket_connaddr = NULL;
	insy_ret->insy_connaddr_len = 0;
	insy_ret->insy_connaddr_alloc = 0;
#endif
	insy_ret->insy_state = INSYTRACKC_INIT;
	return (insytrack_client_t) insy_ret;
}

/* ====================================================================== */

/**
 * Connects the client using the provided settings.
 * @param insy_client The client handle to connect.
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation.
 */
insytrack_client_error_t
insytrack_client_start (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client)
#else
	insy_client)
	insytrack_client_t * insy_client;
#endif
{
	int insy_result;
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_socket_connaddr == NULL )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_ADDRESS_ERROR;
	}
	if ( insy_c->insy_state != INSYTRACKC_INIT )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}

	insy_c->insy_client_socket = socket (insy_c->insy_socket_domain,
		insy_c->insy_socket_type, insy_c->insy_socket_protocol);
	if ( insy_c->insy_client_socket < 0 )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_SOCKET_ERROR;
	}

	insy_result = connect (insy_c->insy_client_socket,
		insy_c->insy_socket_connaddr,
		insy_c->insy_connaddr_len);
	if ( insy_result < 0 )
	{
#ifdef HAVE_CLOSE
		close (insy_c->insy_client_socket);
#endif
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_CONNECT_ERROR;
	}

	insy_c->insy_state = INSYTRACKC_STARTED;
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	return INSYTRACKC_OK;
}

/* ====================================================================== */

/**
 * Sets an address to connect to.
 *	Use before connecting to the server.
 * @param insy_client The client handle to modify.
 * @param insy_addr The address to connect to.
 * @param insy_addrlen The size of the structure pointed to by "addr".
 */
insytrack_client_error_t
insytrack_client_set_addr (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client, struct sockaddr * const insy_addr,
		const socklen_t insy_addrlen)
#else
	insy_client, insy_addr, insy_addrlen)
	insytrack_client_t * insy_client;
	struct sockaddr * const insy_addr;
	const socklen_t insy_addrlen;
#endif
{
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_state != INSYTRACKC_INIT )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}

	if ( insy_c->insy_connaddr_alloc != 0 )
	{
		free (insy_c->insy_socket_connaddr);
		insy_c->insy_connaddr_alloc = 0;
	}

	insy_c->insy_socket_connaddr = insy_addr;
	insy_c->insy_connaddr_len = insy_addrlen;
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket domain to use.
 *	Use before connecting to the server.
 * @param insy_client The client handle to modify.
 * @param insy_domain The socket domain to use.
 */
insytrack_client_error_t
insytrack_client_set_sockdomain (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client, const int insy_domain)
#else
	insy_client, insy_domain)
	insytrack_client_t * insy_client;
	const int insy_domain;
#endif
{
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_state != INSYTRACKC_INIT )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}

	insy_c->insy_socket_domain = insy_domain;
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket type to use.
 *	Use before connecting to the server.
 * @param insy_client The client handle to modify.
 * @param insy_type The socket type to use.
 */
insytrack_client_error_t
insytrack_client_set_socktype (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client, const int insy_type)
#else
	insy_client, insy_type)
	insytrack_client_t * insy_client;
	const int insy_type;
#endif
{
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_state != INSYTRACKC_INIT )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}

	insy_c->insy_socket_type = insy_type;
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket protocol to use.
 *	Use before connecting to the server.
 * @param insy_client The client handle to modify.
 * @param insy_protocol The socket protocol to use.
 */
insytrack_client_error_t
insytrack_client_set_sockproto (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client, const int insy_protocol)
#else
	insy_client, insy_protocol)
	insytrack_client_t * insy_client;
	const int insy_protocol;
#endif
{
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_state != INSYTRACKC_INIT )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}

	insy_c->insy_socket_protocol = insy_protocol;
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets a socket option.
 *	Use after connecting to the server.
 * @param insy_client The client handle to modify.
 * @param level API level, passed as "level" to setsockopt.
 * @param optname the name of the option, passed as "optname" to setsockopt.
 * @param optval the new value of the option, passed as "optval" to setsockopt.
 * @param optlen the size of the option, passed as "optlen" to setsockopt.
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation (INSYTRACKS_PARAM_ERROR = insy_client is NULL,
 *	INSYTRACKS_INVALID_STATE = client not connected,
 *	INSYTRACKS_SOCKET_ERROR = setsockopt error - check errno,
 *	INSYTRACKS_OPERATION_UNAVAILABLE = getsockopt not available on this system).
 */
insytrack_client_error_t
insytrack_client_set_sockopt (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client, int level, int optname,
	const void *optval, socklen_t optlen)
#else
	insy_client, level, optname, optval, optlen)
	insytrack_client_t * insy_client;
	int level;
	int optname;
	const void *optval;
	socklen_t optlen;
#endif
{
#ifdef HAVE_SETSOCKOPT
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;
	int insy_ret;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_state != INSYTRACKC_STARTED )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}
	insy_ret = setsockopt (insy_c->insy_client_socket, level,
		optname, optval, optlen);
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	if ( insy_ret != 0 )
	{
		return INSYTRACKC_SOCKET_ERROR;
	}
	return INSYTRACKC_OK;
#else
	return INSYTRACKC_OPERATION_UNAVAILABLE;
#endif
}

/* ====================================================================== */

/**
 * Gets a socket option.
 *	Use after connecting to the server.
 * @param insy_client The client handle to modify.
 * @param level API level, passed as "level" to getsockopt.
 * @param optname the name of the option, passed as "optname" to getsockopt.
 * @param optval will get the value of the option, passed as "optval" to getsockopt.
 * @param optlen will get the size of the option, passed as "optlen" to getsockopt.
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation (INSYTRACKS_PARAM_ERROR = insy_client is NULL,
 *	INSYTRACKS_INVALID_STATE = client not connected,
 *	INSYTRACKS_SOCKET_ERROR = getsockopt error - check errno,
 *	INSYTRACKS_OPERATION_UNAVAILABLE = getsockopt not available on this system).
 */
insytrack_client_error_t
insytrack_client_get_sockopt (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client, int level, int optname,
	void *optval, socklen_t * optlen)
#else
	insy_client, level, optname, optval, optlen)
	insytrack_client_t * insy_client;
	int level;
	int optname;
	void *optval;
	socklen_t * optlen;
#endif
{
#ifdef HAVE_GETSOCKOPT
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;
	int insy_ret;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	if ( insy_c->insy_state != INSYTRACKC_STARTED )
	{
		pthread_mutex_unlock (&(insy_c->insy_client_mutex));
		return INSYTRACKC_INVALID_STATE;
	}
	insy_ret = getsockopt (insy_c->insy_client_socket, level,
		optname, optval, optlen);
	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	if ( insy_ret != 0 )
	{
		return INSYTRACKC_SOCKET_ERROR;
	}
	return INSYTRACKC_OK;
#else
	return INSYTRACKC_OPERATION_UNAVAILABLE;
#endif
}

/* ====================================================================== */

/**
 * Closes the given client. The handle is destroyed on success.
 * @param insy_client The client handle to close.
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation.
 */
insytrack_client_error_t
insytrack_client_close (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * insy_client)
#else
	insy_client)
	insytrack_client_t * insy_client;
#endif
{
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;

	if ( insy_c == NULL )
	{
		return INSYTRACKC_PARAM_ERROR;
	}

	pthread_mutex_lock (&(insy_c->insy_client_mutex));
	insy_c->insy_state = INSYTRACKC_INIT; /* in case of re-use */
#ifdef HAVE_CLOSE
	if ( insy_c->insy_client_socket != -1 )
	{
		close (insy_c->insy_client_socket);
	}
#endif
	if ( insy_c->insy_connaddr_alloc != 0 )
	{
		free (insy_c->insy_socket_connaddr);
	}

	pthread_mutex_unlock (&(insy_c->insy_client_mutex));
	pthread_mutex_destroy (&(insy_c->insy_client_mutex));
	free (insy_c);
	return INSYTRACKC_OK;
}

