/*
 * InSyTrack - software for tracking program flow, client files.
 *	-- client message management.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_MALLOC_H
# include <malloc.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>	/* write() */
#endif

#include "insytrack_client.h"
#include "libinsytrack_client_init.h"
#include "insytrack_server.h"

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static void libinsytrack_put_number INSYTRACK_PARAMS ((
	unsigned char * const insy_buf,
	const unsigned int insy_field_len,
	const unsigned int insy_value));
#endif

/**
 * Puts a number in the given buffer.
 * @param insy_buf the buffer to put the number in.
 * @param insy_field_len The length of the field containing
 *	the number (in bytes).
 * @param insy_value the value to put in the buffer.
 */
static void libinsytrack_put_number (
#ifdef INSYTRACK_ANSIC
	unsigned char * const insy_buf,
	const unsigned int insy_field_len,
	const unsigned int insy_value)
#else
	insy_buf, insy_field_len, insy_value)
	unsigned char * const insy_buf;
	const unsigned int insy_field_len;
	const unsigned int insy_value;
#endif
{
	unsigned int i;

	if ( (insy_buf == NULL) || (insy_field_len == 0) )
	{
		return;
	}
	/* extract the number from the byte buffer */
	for ( i = 0; i < insy_field_len; i++ )
	{
		insy_buf[i] = (unsigned char) ((insy_value
			>> ((insy_field_len - i - 1)*8)) & 0x0FF);
	}
}

/* ====================================================================== */

/**
 * Creates a new InSyTrack message.
 * @return a new InSyTrack message
 *	or INSYTRACK_MESSAGE_INVALID in case of error.
 */
insytrack_message_t *
insytrack_msg_create (
#ifdef INSYTRACK_ANSIC
	void
#endif
)
{
	insytrack_message_t * insy_ret;

	insy_ret = (insytrack_message_t *) malloc (sizeof (insytrack_message_t));
	if ( insy_ret == NULL )
	{
		return INSYTRACK_MESSAGE_INVALID;
	}
	insy_ret->insy_type.insy_type = INSYTRACK_MESSAGE_DATA;
	insy_ret->insy_msgid_len.insy_msgid_len_int = 0;
	insy_ret->insy_msgbody_len.insy_msgbody_len_int = 0;
	insy_ret->insy_msgid = NULL;
	insy_ret->insy_msgbody = NULL;
	return insy_ret;
}

/* ====================================================================== */

/**
 * Destroys an InSyTrack message.
 * @param insy_msg the InSyTrack message to destroy.
 */
void
insytrack_msg_destroy (
#ifdef INSYTRACK_ANSIC
	insytrack_message_t * insy_msg)
#else
	insy_msg)
	insytrack_message_t * insy_msg;
#endif
{
	if ( insy_msg == NULL )
	{
		return;
	}
	if ( insy_msg->insy_msgid != NULL )
	{
		free (insy_msg->insy_msgid);
	}
	if ( insy_msg->insy_msgbody != NULL )
	{
		free (insy_msg->insy_msgbody);
	}
	free (insy_msg);
}

/* ====================================================================== */

/**
 * Sends an InSyTrack message.
 * @param insy_client The client handle to send the message with.
 * @param insy_msg the InSyTrack message to send.
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation.
 */
insytrack_client_error_t
insytrack_msg_send (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * const insy_client,
	const insytrack_message_t * const insy_msg)
#else
	insy_client, insy_msg)
	insytrack_client_t * const insy_client;
	const insytrack_message_t * const insy_msg;
#endif
{
	int insy_result;
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;
	unsigned char insy_msgparams[INSYTRACK_MSG_MARK_LEN
		+ INSYTRACK_MSG_TYPE_LEN
		+ INSYTRACK_MSGID_LENGTH_LEN
		+ INSYTRACK_MSGBODY_LENGTH_LEN];
	int i;
	unsigned int insy_msgid_len;
	unsigned int insy_msgbody_len;

	if ( (insy_msg == NULL) || (insy_c == NULL) )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	if ( insy_c->insy_client_socket < 0 )
	{
		return INSYTRACKC_SOCKET_ERROR;
	}

	if ( insy_msg->insy_msgid == NULL )
	{
		insy_msgid_len = 0;
	}
	else
	{
		insy_msgid_len = insy_msg->insy_msgid_len.insy_msgid_len_int;
	}
	if ( insy_msg->insy_msgbody == NULL )
	{
		insy_msgbody_len = 0;
	}
	else
	{
		insy_msgbody_len = insy_msg->insy_msgbody_len.insy_msgbody_len_int;
	}
	i = 0;
	libinsytrack_put_number (&(insy_msgparams[i]), INSYTRACK_MSG_MARK_LEN,
		INSYTRACK_MSG_MARK);
	i += INSYTRACK_MSG_MARK_LEN;
	libinsytrack_put_number (&(insy_msgparams[i]), INSYTRACK_MSG_TYPE_LEN,
		insy_msg->insy_type.insy_type_int);
	i += INSYTRACK_MSG_TYPE_LEN;
	libinsytrack_put_number (&(insy_msgparams[i]), INSYTRACK_MSGID_LENGTH_LEN,
		insy_msgid_len);
	i += INSYTRACK_MSGID_LENGTH_LEN;
	libinsytrack_put_number (&(insy_msgparams[i]), INSYTRACK_MSGBODY_LENGTH_LEN,
		insy_msgbody_len);
	i += INSYTRACK_MSGBODY_LENGTH_LEN;

	insy_result = write (insy_c->insy_client_socket, insy_msgparams,
		(size_t)i);
	if ( (insy_result < 0) || (insy_result != i) )
	{
		return INSYTRACKC_SEND_ERROR;
	}
	if ( (insy_msg->insy_msgid != NULL)
		&& (insy_msg->insy_msgid_len.insy_msgid_len_int > 0) )
	{
		insy_result = write (insy_c->insy_client_socket,
			insy_msg->insy_msgid,
			insy_msg->insy_msgid_len.insy_msgid_len_int);
		if ( (insy_result < 0) ||
			((unsigned int)insy_result != insy_msg->insy_msgid_len.insy_msgid_len_int) )
		{
			return INSYTRACKC_SEND_ERROR;
		}
	}
	if ( (insy_msg->insy_msgbody != NULL)
		&& (insy_msg->insy_msgbody_len.insy_msgbody_len_int > 0) )
	{
		insy_result = write (insy_c->insy_client_socket,
			insy_msg->insy_msgbody,
			insy_msg->insy_msgbody_len.insy_msgbody_len_int);
		if ( (insy_result < 0) ||
			((unsigned int)insy_result != insy_msg->insy_msgbody_len.insy_msgbody_len_int) )
		{
			return INSYTRACKC_SEND_ERROR;
		}
	}
	return INSYTRACKC_OK;
}

/* ====================================================================== */

/**
 * Creates and sends a message with the given ID and type.
 * @param insy_client The client handle to send the mark with.
 * @param insy_id The message ID to mark. Must NOT be NULL.
 * @param insy_id_len The length message ID to mark.
 * @param insy_data Any additional data to send with the mark
 *	(like: filename, location etc.). Can be NULL.
 * @param insy_data_len The length of any additional data
 *	to send with the mark.
 * @param insy_type The type of the message.
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation.
 */
insytrack_client_error_t
insytrack_create_and_send (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * const insy_client, void * insy_id,
	unsigned short int insy_id_len, void * insy_data,
	unsigned short int insy_data_len, const insytrack_message_type_t insy_type)
#else
	insy_client, insy_id, insy_id_len, insy_data, insy_data_len, insy_type)
	insytrack_client_t * const insy_client;
	void * insy_id;
	unsigned short int insy_id_len;
	void * insy_data;
	unsigned short int insy_data_len;
	const insytrack_message_type_t insy_type;
#endif
{
	struct insytrack_client * insy_c = (struct insytrack_client *) insy_client;
	insytrack_message_t * insy_msg;
	insytrack_client_error_t insy_send_ret;

	if ( (insy_id == NULL) || (insy_c == NULL) )
	{
		return INSYTRACKC_PARAM_ERROR;
	}
	if ( insy_c->insy_client_socket < 0 )
	{
		return INSYTRACKC_SOCKET_ERROR;
	}

	if ( insy_id == NULL )
	{
		insy_id_len = 0;
	}
	if ( insy_data == NULL )
	{
		insy_data_len = 0;
	}
	insy_msg = insytrack_msg_create ();
	if ( insy_msg == INSYTRACK_MESSAGE_INVALID )
	{
		return INSYTRACKC_MALLOC_ERROR;
	}
	insy_msg->insy_type = insy_type;
	insy_msg->insy_msgid_len.insy_msgid_len_int = insy_id_len;
	insy_msg->insy_msgbody_len.insy_msgbody_len_int = insy_data_len;
	insy_msg->insy_msgid = insy_id;
	insy_msg->insy_msgbody = insy_data;
	insy_send_ret = insytrack_msg_send (insy_client, insy_msg);
	/* this is user-provided data, don't free() it */
	insy_msg->insy_msgid = NULL;
	insy_msg->insy_msgbody = NULL;
	insytrack_msg_destroy (insy_msg);
	return insy_send_ret;
}

/* ====================================================================== */

/**
 * Marks a usage of the given ID. This is the key feature - it tells the
 *	server that new flow tracking data is available. Should be called
 *	at every place the ID is supposed to be tracked (ID should not change).
 * @param insy_client The client handle to send the mark with.
 * @param insy_id the message ID to mark. Must NOT be NULL.
 * @param insy_id_len the length message ID to mark.
 * @param insy_data any additional data to send with the mark
 *	(like: filename, location etc.). Can be NULL.
 * @param insy_data_len the length of any additional data to send with the mark
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation.
 */
insytrack_client_error_t
insytrack_mark_id (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * const insy_client, void * insy_id,
	unsigned short int insy_id_len, void * insy_data,
	unsigned short int insy_data_len)
#else
	insy_client, insy_id, insy_id_len, insy_data, insy_data_len)
	insytrack_client_t * const insy_client;
	void * insy_id;
	unsigned short int insy_id_len;
	void * insy_data;
	unsigned short int insy_data_len;
#endif
{
	insytrack_message_type_t insy_type;

	insy_type.insy_type = INSYTRACK_MESSAGE_DATA;
	return insytrack_create_and_send (insy_client, insy_id, insy_id_len,
		insy_data, insy_data_len, insy_type);
}

/* ====================================================================== */

/**
 * Marks the last usage of the given ID. It tells the server that the flow
 *	connected with the given ID has been finished. Should be called
 *	at closing the flow (ID should not change).
 * @param insy_client The client handle to send the mark with.
 * @param insy_id the message ID to mark. Must NOT be NULL.
 * @param insy_id_len the length message ID to mark.
 * @param insy_data any additional data to send with the mark
 *	(like: filename, location etc.). Can be NULL.
 * @param insy_data_len the length of any additional data to send with the mark
 * @return an insytrack_client_error_t value specifying the
 *	status of the operation.
 */
insytrack_client_error_t
insytrack_mark_id_last (
#ifdef INSYTRACK_ANSIC
	insytrack_client_t * const insy_client, void * insy_id,
	unsigned short int insy_id_len, void * insy_data,
	unsigned short int insy_data_len)
#else
	insy_client, insy_id, insy_id_len, insy_data, insy_data_len)
	insytrack_client_t * const insy_client;
	void * insy_id;
	unsigned short int insy_id_len;
	void * insy_data;
	unsigned short int insy_data_len;
#endif
{
	insytrack_message_type_t insy_type;

	insy_type.insy_type = INSYTRACK_MESSAGE_LAST;
	return insytrack_create_and_send (insy_client, insy_id, insy_id_len,
		insy_data, insy_data_len, insy_type);
}

