/*
 * InSyTrack - software for tracking program flow, server files.
 *	-- server initialization & shutdown.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>	/* htonl(), htons() */
#endif

#ifdef HAVE_ARPA_INET_H
# include <arpa/inet.h>		/* htonl(), htons() */
#endif

/* time headers for select() (the old way) */
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  ifdef HAVE_TIME_H
#   include <time.h>
#  endif
# endif
#endif

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_MALLOC_H
# include <malloc.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>	/* select() (the old way), read(), close() */
#endif

/* select () - the new way */
#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "insytrack_server.h"
#include "libinsytrack_server_init.h"
#include "libinsytrack_server_thread.h"
#include "libinsytrack_server_tlist.h"

static void * insytrack_dispatcher INSYTRACK_PARAMS ((void * param));

/* ====================================================================== */

/**
 * Creates a new server.
 * @return a server handle with the default configuration
 *	or INSYTRACK_SERVER_INVALID in case of error.
 */
insytrack_server_t
insytrack_server_create (
#ifdef INSYTRACK_ANSIC
	void
#endif
)
{
	int insy_result;
	struct insytrack_server * insy_ret;

	insy_ret = (struct insytrack_server *) malloc (sizeof (struct insytrack_server));
	if ( insy_ret == NULL )
	{
		return INSYTRACK_SERVER_INVALID;
	}
	insy_result = pthread_mutex_init (&(insy_ret->insy_server_mutex), NULL);
	if ( insy_result != 0 )
	{
		free (insy_ret);
		return INSYTRACK_SERVER_INVALID;
	}
	insy_ret->insy_server_socket = -1;
	insy_ret->insy_socket_domain = INSYTRACK_SERVER_SOCKET_DOMAIN;
	insy_ret->insy_socket_type = INSYTRACK_SERVER_SOCKET_TYPE;
	insy_ret->insy_socket_protocol = INSYTRACK_SERVER_SOCKET_PROTOCOL;
#ifdef HAVE_NETINET_IN_H
	insy_ret->insy_socket_bindaddr =
		(struct sockaddr *) malloc (sizeof (struct sockaddr_in));
	if ( insy_ret->insy_socket_bindaddr != NULL )
	{
		((struct sockaddr_in *)(insy_ret->insy_socket_bindaddr))->sin_family
			= AF_INET;
		((struct sockaddr_in *)(insy_ret->insy_socket_bindaddr))->sin_port
			= htons (INSYTRACK_SERVER_SOCKET_PORT); /* warnings inside */
		((struct sockaddr_in *)(insy_ret->insy_socket_bindaddr))->sin_addr.s_addr
			= htonl (INADDR_ANY); /* warnings inside */
		insy_ret->insy_bindaddr_len = sizeof (struct sockaddr_in);
		insy_ret->insy_bindaddr_alloc = 1;
	}
	else
	{
		insy_ret->insy_bindaddr_len = 0;
		insy_ret->insy_bindaddr_alloc = 0;
	}
#else
	insy_ret->insy_socket_bindaddr = NULL;
	insy_ret->insy_bindaddr_len = 0;
	insy_ret->insy_bindaddr_alloc = 0;
#endif
	insy_ret->insy_backlog = INSYTRACK_SERVER_SOCKET_BACKLOG;
	insy_ret->insy_msgrecvd_callback = NULL;
	insy_ret->insy_state = INSYTRACKS_INIT;
	return (insytrack_server_t) insy_ret;
}

/* ====================================================================== */

/**
 * Starts the server with the provided settings.
 * @param insy_server The server handle to start.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_start (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server)
#else
	insy_server)
	insytrack_server_t * insy_server;
#endif
{
	int insy_result;
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_socket_bindaddr == NULL )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_ADDRESS_ERROR;
	}
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_server_socket = socket (insy_s->insy_socket_domain,
		insy_s->insy_socket_type, insy_s->insy_socket_protocol);
	if ( insy_s->insy_server_socket < 0 )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_SOCKET_ERROR;
	}

	insy_result = bind (insy_s->insy_server_socket,
		insy_s->insy_socket_bindaddr,
		insy_s->insy_bindaddr_len);
	if ( insy_result < 0 )
	{
#ifdef HAVE_CLOSE
		close (insy_s->insy_server_socket);
#endif
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_BIND_ERROR;
	}

#ifdef HAVE_ERRNO_H
	errno = 0;
#endif
	insy_result = listen (insy_s->insy_server_socket, insy_s->insy_backlog);
	if ( insy_result < 0 )
	{
		/* Some sockets don't support the listen() operation. Check this
		before failing. If this such a socket, proceed despite the error. */
#ifdef HAVE_ERRNO_H
		if ( errno != EOPNOTSUPP )
#endif
		{
#ifdef HAVE_CLOSE
			close (insy_s->insy_server_socket);
#endif
			pthread_mutex_unlock (&(insy_s->insy_server_mutex));
			return INSYTRACKS_LISTEN_ERROR;
		}
	}

	insy_result = pthread_mutex_init (&(insy_s->insy_list_mutex), NULL);
	if ( insy_result != 0 )
	{
#ifdef HAVE_CLOSE
		close (insy_s->insy_server_socket);
#endif
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_MUTEX_ERROR;
	}
	insy_s->insy_shutting_down = 0;
	insy_s->insy_tlist_head = NULL;

	insy_result = pthread_create (&(insy_s->insy_dispatch_thread), NULL,
		&insytrack_dispatcher, insy_s);
	if ( insy_result != 0 )
	{
#ifdef HAVE_CLOSE
		close (insy_s->insy_server_socket);
#endif
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_THREAD_ERROR;
	}

	insy_s->insy_state = INSYTRACKS_STARTED;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets an address to bind to.
 *	Use before starting the server.
 * @param insy_server The server handle to modify.
 * @param insy_addr The address to bind to.
 * @param insy_addrlen The size of the structure pointed to by "addr".
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_addr (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, struct sockaddr * const insy_addr,
		const socklen_t insy_addrlen)
#else
	insy_server, insy_addr, insy_addrlen)
	insytrack_server_t * insy_server;
	struct sockaddr * const insy_addr;
	const socklen_t insy_addrlen;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	if ( insy_s->insy_bindaddr_alloc != 0 )
	{
		free (insy_s->insy_socket_bindaddr);
		insy_s->insy_bindaddr_alloc = 0;
	}

	insy_s->insy_socket_bindaddr = insy_addr;
	insy_s->insy_bindaddr_len = insy_addrlen;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket domain to use.
 *	Use before starting the server.
 * @param insy_server The server handle to modify.
 * @param insy_domain The socket domain to use.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_sockdomain (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, const int insy_domain)
#else
	insy_server, insy_domain)
	insytrack_server_t * insy_server;
	const int insy_domain;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_socket_domain = insy_domain;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket type to use.
 *	Use before starting the server.
 * @param insy_server The server handle to modify.
 * @param insy_type The socket type to use.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_socktype (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, const int insy_type)
#else
	insy_server, insy_type)
	insytrack_server_t * insy_server;
	const int insy_type;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_socket_type = insy_type;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket protocol to use.
 *	Use before starting the server.
 * @param insy_server The server handle to modify.
 * @param insy_protocol The socket protocol to use.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_sockproto (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, const int insy_protocol)
#else
	insy_server, insy_protocol)
	insytrack_server_t * insy_server;
	const int insy_protocol;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_socket_protocol = insy_protocol;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the socket backlog size (the maximum number of waiting connections).
 *	Use before starting the server.
 * @param insy_server The server handle to modify.
 * @param insy_backlog_size The socket insy_backlog size.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_backlog (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, const int insy_backlog_size)
#else
	insy_server, insy_backlog_size)
	insytrack_server_t * insy_server;
	const int insy_backlog_size;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_backlog = insy_backlog_size;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets a socket option of the server socket.
 *	Use after starting the server.
 * @param insy_server The server handle to modify.
 * @param level API level, passed as "level" to setsockopt.
 * @param optname the name of the option, passed as "optname" to setsockopt.
 * @param optval the new value of the option, passed as "optval" to setsockopt.
 * @param optlen the size of the option, passed as "optlen" to setsockopt.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation (INSYTRACKS_PARAM_ERROR = insy_server is NULL,
 *	INSYTRACKS_INVALID_STATE = server not started,
 *	INSYTRACKS_SOCKET_ERROR = setsockopt error - check errno,
 *	INSYTRACKS_OPERATION_UNAVAILABLE = getsockopt not available on this system).
 */
insytrack_server_error_t
insytrack_server_set_sockopt (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, int level, int optname,
	const void *optval, socklen_t optlen)
#else
	insy_server, level, optname, optval, optlen)
	insytrack_server_t * insy_server;
	int level;
	int optname;
	const void *optval;
	socklen_t optlen;
#endif
{
#ifdef HAVE_SETSOCKOPT
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;
	int insy_ret;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_STARTED )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}
	insy_ret = setsockopt (insy_s->insy_server_socket, level,
		optname, optval, optlen);
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	if ( insy_ret != 0 )
	{
		return INSYTRACKS_SOCKET_ERROR;
	}
	return INSYTRACKS_OK;
#else
	return INSYTRACKS_OPERATION_UNAVAILABLE;
#endif
}

/* ====================================================================== */

/**
 * Gets a socket option of the server socket.
 *	Use after starting the server.
 * @param insy_server The server handle to modify.
 * @param level API level, passed as "level" to getsockopt.
 * @param optname the name of the option, passed as "optname" to getsockopt.
 * @param optval will get the value of the option, passed as "optval" to getsockopt.
 * @param optlen will get the size of the option, passed as "optlen" to getsockopt.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation (INSYTRACKS_PARAM_ERROR = insy_server is NULL,
 *	INSYTRACKS_INVALID_STATE = server not started,
 *	INSYTRACKS_SOCKET_ERROR = getsockopt error - check errno,
 *	INSYTRACKS_OPERATION_UNAVAILABLE = getsockopt not available on this system).
 */
insytrack_server_error_t
insytrack_server_get_sockopt (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server, int level, int optname,
	void *optval, socklen_t * optlen)
#else
	insy_server, level, optname, optval, optlen)
	insytrack_server_t * insy_server;
	int level;
	int optname;
	void *optval;
	socklen_t * optlen;
#endif
{
#ifdef HAVE_GETSOCKOPT
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;
	int insy_ret;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_STARTED )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}
	insy_ret = getsockopt (insy_s->insy_server_socket, level,
		optname, optval, optlen);
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	if ( insy_ret != 0 )
	{
		return INSYTRACKS_SOCKET_ERROR;
	}
	return INSYTRACKS_OK;
#else
	return INSYTRACKS_OPERATION_UNAVAILABLE;
#endif
}

/* ====================================================================== */

/**
 * Sets the "message received" callback function.
 *	Use before starting the server.
 * @param insy_server The server to set the callback on.
 * @param insy_callback The callback to register (NULL to un-register).
 * @param insy_data The data passed to the callback.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_msgrcvd_fcn (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server,
	const insytrack_msg_recvd_fun_t insy_callback,
	void * insy_data)
#else
	insy_server, insy_callback, insy_data)
	insytrack_server_t * insy_server;
	const insytrack_msg_recvd_fun_t insy_callback;
	void * insy_data;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_msgrecvd_callback = insy_callback;
	insy_s->insy_msgrecvd_callback_data = insy_data;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the "client connected" callback function.
 *	Use before starting the server.
 * @param insy_server The server to set the callback on.
 * @param insy_callback The callback to register (NULL to un-register).
 * @param insy_data The data passed to the callback.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_conn_fcn (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server,
	const insytrack_client_connected_fun_t insy_callback,
	void * insy_data)
#else
	insy_server, insy_callback, insy_data)
	insytrack_server_t * insy_server;
	const insytrack_client_connected_fun_t insy_callback;
	void * insy_data;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_conn_callback = insy_callback;
	insy_s->insy_conn_callback_data = insy_data;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Sets the "client disconnected" callback function.
 *	Use before starting the server.
 * @param insy_server The server to set the callback on.
 * @param insy_callback The callback to register (NULL to un-register).
 * @param insy_data The data passed to the callback.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_set_disconn_fcn (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server,
	const insytrack_client_disconnected_fun_t insy_callback,
	void * insy_data)
#else
	insy_server, insy_callback, insy_data)
	insytrack_server_t * insy_server;
	const insytrack_client_disconnected_fun_t insy_callback;
	void * insy_data;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}
	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	if ( insy_s->insy_state != INSYTRACKS_INIT )
	{
		pthread_mutex_unlock (&(insy_s->insy_server_mutex));
		return INSYTRACKS_INVALID_STATE;
	}

	insy_s->insy_disconn_callback = insy_callback;
	insy_s->insy_disconn_callback_data = insy_data;
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * Closes the given server. The handle is destroyed on success.
 * @param insy_server The server handle to close.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t
insytrack_server_close (
#ifdef INSYTRACK_ANSIC
	insytrack_server_t * insy_server)
#else
	insy_server)
	insytrack_server_t * insy_server;
#endif
{
	struct insytrack_server * insy_s = (struct insytrack_server *) insy_server;

	if ( insy_s == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}

	pthread_mutex_lock (&(insy_s->insy_server_mutex));
	insy_s->insy_state = INSYTRACKS_INIT; /* in case of re-use */
	insy_s->insy_shutting_down = 1;
#ifdef HAVE_CLOSE
	if ( insy_s->insy_server_socket != -1 )
	{
		close (insy_s->insy_server_socket);
	}
#endif
	pthread_join (insy_s->insy_dispatch_thread, NULL);
	pthread_mutex_destroy (&(insy_s->insy_list_mutex));
	if ( insy_s->insy_bindaddr_alloc != 0 )
	{
		free (insy_s->insy_socket_bindaddr);
	}
	pthread_mutex_unlock (&(insy_s->insy_server_mutex));
	pthread_mutex_destroy (&(insy_s->insy_server_mutex));
	free (insy_s);
	return INSYTRACKS_OK;
}

/* ====================================================================== */

/**
 * The connection dispatcher.
 * @param insy_param The parameter passed by pthread_create().
 * @return NULL.
 */
static void *
insytrack_dispatcher (
#ifdef INSYTRACK_ANSIC
	void * insy_param)
#else
	insy_param)
	void * insy_param;
#endif
{
	int insy_client_socket;
	struct insytrack_server * insy_server;
	insytrack_server_error_t insy_result;
	struct insytrack_worker_data insy_twd;
	int insy_all_closed;
	struct insytrack_thread_list * insy_curr_elem;
	int insy_thread_started = 0;
	struct sockaddr * insy_client_addr = NULL;
	socklen_t insy_server_addr_len;
#ifdef INSYTRACK_HAVE_SELECT
	struct timeval insy_tv;
	fd_set insy_set;
#endif

	if ( insy_param == NULL )
	{
		return NULL;
	}

	insy_server = (struct insytrack_server *) insy_param;
	if ( insy_server->insy_bindaddr_len > 0 )
	{
		insy_client_addr = (struct sockaddr *)
			malloc (insy_server->insy_bindaddr_len);
	}

	do
	{
#ifdef INSYTRACK_HAVE_SELECT
		/* select() can destroy its parameters, so reset every time */
		FD_ZERO (&insy_set);
		FD_SET (insy_server->insy_server_socket, &insy_set);	/* warnings are inside this macro */
		insy_tv.tv_sec = INSYTRACK_MAX_SELECT_SECONDS;
		insy_tv.tv_usec = 0;
		if ( select (insy_server->insy_server_socket+1,
			&insy_set, NULL, NULL, &insy_tv) > 0 )
		{
#endif
#ifdef HAVE_ERRNO_H
			errno = 0;
#endif
			insy_server_addr_len = insy_server->insy_bindaddr_len;
			insy_client_socket = accept (
				insy_server->insy_server_socket,
				insy_client_addr, &insy_server_addr_len);
#ifdef INSYTRACK_HAVE_SELECT
		}
		else
		{
			insy_client_socket = -1; /* marker allowing exiting the loop */
		}
#endif
		if ( insy_client_socket < 0 )
		{
#ifdef HAVE_ERRNO_H
			if ( errno != EOPNOTSUPP )
#endif
			{
				/* ignore the error until the server is notified to shutdown. */
				continue;
			}
#ifdef HAVE_ERRNO_H
			else
			{
				if ( insy_thread_started != 0 )
				{
					continue;
				}
				insy_thread_started = 1;
				insy_twd.insy_client_socket = insy_server->insy_server_socket;
			}
#endif
		}
		else
		{
			insy_twd.insy_client_socket = insy_client_socket;
		}
		if ( insy_server->insy_conn_callback != NULL )
		{
			(*(insy_server->insy_conn_callback)) (insy_client_addr,
				insy_server_addr_len,
				insy_server->insy_conn_callback_data);
		}
		insy_twd.insy_server = insy_server;
		insy_twd.insy_client_addr = insy_client_addr;
		insy_twd.insy_client_addr_len = insy_server_addr_len;
		insy_result = insytrack_create_worker (insy_server, &insy_twd);
		if ( insy_result != INSYTRACKS_OK )
		{
			insy_thread_started = 0;
#ifdef HAVE_CLOSE
			close (insy_client_socket);
#endif
		}

	} while (insy_server->insy_shutting_down == 0);

	/* wait for all the threads: */
#ifdef HAVE_SLEEP
	sleep (2*INSYTRACK_MAX_SELECT_SECONDS);
#else
# ifdef INSYTRACK_HAVE_SELECT
	/* select() can destroy its parameters, so reset every time */
	FD_ZERO (&insy_set);
	insy_tv.tv_sec = 2*INSYTRACK_MAX_SELECT_SECONDS;
	insy_tv.tv_usec = 0;
	select (0, &insy_set, NULL, NULL, &insy_tv);
# else
	{
		volatile int i;
		for ( i = 0; i < (1<<30); i++ );
	}
# endif
#endif
	insy_all_closed = 0;
	while ( insy_all_closed == 0 )
	{
#ifdef HAVE_SLEEP
		sleep (INSYTRACK_MAX_SELECT_SECONDS);
#else
# ifdef INSYTRACK_HAVE_SELECT
		/* select() can destroy its parameters, so reset every time */
		FD_ZERO (&insy_set);
		insy_tv.tv_sec = INSYTRACK_MAX_SELECT_SECONDS;
		insy_tv.tv_usec = 0;
		select (0, &insy_set, NULL, NULL, &insy_tv);
# else
		volatile int i;
		for ( i = 0; i < (1<<30); i++ );
# endif
#endif
		pthread_mutex_lock (&(insy_server->insy_list_mutex));
		if ( insy_server->insy_tlist_head == NULL )
		{
			insy_all_closed = 1;
		}
		else
		{
			insy_curr_elem = insy_server->insy_tlist_head;
			while ( insy_curr_elem != NULL )
			{
				if ( insy_curr_elem->insy_thread_data != NULL )
				{
					if ( insy_curr_elem->insy_thread_data->insy_is_running == 0 )
					{
						/* thread didn't start, join with it */
						pthread_join (insy_curr_elem->insy_thread_id, NULL);
						insytrack_remove_worker (insy_server,
							insy_curr_elem->insy_thread_id);
						if (insy_server->insy_tlist_head == NULL)
						{
							insy_all_closed = 1;
						}
					}
				}
			}
		}
		pthread_mutex_unlock (&(insy_server->insy_list_mutex));
	}

	if ( insy_client_addr != NULL )
	{
		free (insy_client_addr);
	}

	return NULL;
}

