/*
 * InSyTrack - software for tracking program flow, server files.
 *	-- server worker threads' file.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#include <stdio.h>

#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

/* time headers for select() (the old way) */
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  ifdef HAVE_TIME_H
#   include <time.h>
#  endif
# endif
#endif

#ifdef HAVE_STRING_H
# if ((!defined STDC_HEADERS) || (!STDC_HEADERS)) && (defined HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>	/* memset() */
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>	/* select() (the old way), read(), close() */
#endif

/* select () - the new way */
#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_MALLOC_H
# include <malloc.h>
#endif

#include "insytrack_server.h"
#include "libinsytrack_server_init.h"
#include "libinsytrack_server_thread.h"
#include "libinsytrack_server_tlist.h"

/**
 * The buffer size for each read, in bytes.
 */
#define INSYTRACK_MSG_BUF_SIZE 1000000

/**
 * The enumeration of possible return values when reading messages.
 */
enum insytrack_msgret
{
	/**
	 * The return value when a messages has been returned.
	 */
	INSYTRACK_MSGRET_SUCCESS,

	/**
	 * The return value when message reading should be aborted.
	 */
	INSYTRACK_MSGRET_ABORT,

	/**
	 * The return value when a bad parameter was passed to message reading.
	 */
	INSYTRACK_MSGRET_BADPARAM
};

/**
 * The type of possible return values when reading messages.
 */
typedef enum insytrack_msgret insytrack_msgret_t;

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static unsigned int libinsytrack_extract_number INSYTRACK_PARAMS ((
	const unsigned char * const insy_buf,
	const unsigned int insy_field_len));
#endif

/**
 * Reads a number from the given buffer.
 * @param insy_buf the buffer to read the number from.
 * @param insy_field_len The length of the field containing
 *	the number (in bytes).
 * @return the value read from the buffer
 */
static unsigned int libinsytrack_extract_number (
#ifdef INSYTRACK_ANSIC
	const unsigned char * const insy_buf,
	const unsigned int insy_field_len)
#else
	insy_buf, insy_field_len)
	const unsigned char * const insy_buf;
	const unsigned int insy_field_len;
#endif
{
	unsigned int i;
	unsigned int insy_val = 0xFFFFFFFF;

	if ( (insy_buf == NULL) || (insy_field_len == 0) )
	{
		return insy_val;
	}
	/* extract the number from the byte buffer */
	for ( i = 0; i < insy_field_len; i++ )
	{
		insy_val = (insy_val << 8) | (insy_buf[i] & 0x0FF);
	}
	/* clear the unnecessary bits: */
	if ( insy_field_len < sizeof (unsigned int) / sizeof (char) )
	{
		insy_val = (unsigned int)(insy_val
			& ((unsigned int)(1 << (insy_field_len * 8)) - 1));
	}

	return insy_val;
}

/* ======================================================================== */

#ifndef INSYTRACK_ANSIC
static unsigned int libinsytrack_find_marker INSYTRACK_PARAMS ((
	const unsigned char * const insy_buf,
	const unsigned int insy_buf_len));
#endif

/**
 * Finds the InSyTrack message start marker in the given buffer.
 * @param insy_buf the buffer to find he marker in.
 * @param insy_buf_len The length the buffer (in bytes).
 * @return the position of the marker of 0xFFFFFFFF if not found.
 */
static unsigned int libinsytrack_find_marker (
#ifdef INSYTRACK_ANSIC
	const unsigned char * const insy_buf,
	const unsigned int insy_buf_len)
#else
	insy_buf, insy_buf_len)
	const unsigned char * const insy_buf;
	const unsigned int insy_buf_len;
#endif
{
	unsigned int i;
	unsigned int j;
	unsigned int insy_position = 0xFFFFFFFF;

	if ( (insy_buf == NULL) || (insy_buf_len == 0) )
	{
		return insy_position;
	}
	for ( i = 0; i < insy_buf_len - INSYTRACK_MSG_MARK_LEN; i++ )
	{
		for ( j = 0; j < INSYTRACK_MSG_MARK_LEN; j++ )
		{
			if ( insy_buf[i+j] != ((INSYTRACK_MSG_MARK >>
				(INSYTRACK_MSG_MARK_LEN - (j+1))*8) & 0x0FF) )
			{
				break;
			}
		}
		if ( j < INSYTRACK_MSG_MARK_LEN )
		{
			return i;
		}
	}

	return insy_position;
}

/* ======================================================================== */


#ifndef INSYTRACK_ANSIC
static insytrack_msgret_t libinsytrack_server_read_message
	INSYTRACK_PARAMS ((struct insytrack_worker_data * const insy_twd,
	insytrack_message_t * const insy_msg));
#endif

/**
 * Reads a message from the given file descriptor and puts it in the buffer.
 *	Waits at most INSYTRACK_MAX_SELECT_SECONDS seconds for data.
 * @param insy_twd The thread's working data.
 * @param insy_msg The place to return the received message in.
 * @return an insytrack_msgret_t value specifying the status of the opertions.
 */
static insytrack_msgret_t libinsytrack_server_read_message (
#ifdef INSYTRACK_ANSIC
	struct insytrack_worker_data * const insy_twd,
	insytrack_message_t * const insy_msg)
#else
	insy_twd, insy_msg)
	struct insytrack_worker_data * const insy_twd;
	insytrack_message_t * const insy_msg;
#endif
{
	int insy_bytes_read = -1;
	unsigned int insy_msg_length = 0xFFFFFFFF;
	unsigned int insy_msg_type = 0xFFFFFFFF;
	unsigned int insy_msgid_length = 0xFFFFFFFF;
	unsigned int insy_got_msg_mark = 0;
	unsigned int insy_total_length;
	unsigned int insy_mark_position;
	insytrack_msgret_t insy_res;
#ifdef INSYTRACK_HAVE_SELECT
	struct timeval insy_tv;
	fd_set insy_set_r;
	fd_set insy_set_ex;
#endif
#if (!defined HAVE_MEMMOVE) || (!defined HAVE_MEMCPY)
	size_t insy_offset;
#endif

	if ( (insy_twd == NULL) || (insy_msg == NULL) )
	{
		return INSYTRACK_MSGRET_BADPARAM;
	}
	if ( (insy_twd->insy_server == NULL)
		|| (insy_twd->insy_client_socket < 0)
		|| (insy_twd->insy_buf == NULL) )
	{
		return INSYTRACK_MSGRET_BADPARAM;
	}

	insy_res = INSYTRACK_MSGRET_ABORT;
	do
	{
		/* First, check how many bytes we have in the buffer.
		If we have a full message, don't wait for the select/read,
		process it immediately. */
		if ( (insy_twd->insy_currlen >= INSYTRACK_MSG_MARK_LEN)
			&& (insy_got_msg_mark == 0))
		{
			insy_got_msg_mark = libinsytrack_extract_number (
				&(insy_twd->insy_buf[insy_twd->insy_bufhead]),
				INSYTRACK_MSG_MARK_LEN);
			if ( insy_got_msg_mark != INSYTRACK_MSG_MARK )
			{
				/* garbage in the socket detected */
				insy_mark_position = libinsytrack_find_marker (
					&(insy_twd->insy_buf[insy_twd->insy_bufhead]),
					insy_twd->insy_currlen);
				if ( insy_mark_position != 0xFFFFFFFF )
				{
					insy_twd->insy_bufhead = insy_mark_position;
				}
				else
				{
					/* marker not found - assume whole buffer useless,
					but keep the last 3 bytes in case they may contain the
					start of the marker */
					insy_twd->insy_buf[0] = insy_twd->insy_buf[
						insy_twd->insy_bufhead + insy_twd->insy_currlen-3];
					insy_twd->insy_buf[1] = insy_twd->insy_buf[
						insy_twd->insy_bufhead + insy_twd->insy_currlen-2];
					insy_twd->insy_buf[2] = insy_twd->insy_buf[
						insy_twd->insy_bufhead + insy_twd->insy_currlen-1];
					insy_twd->insy_currlen = 3;
					insy_twd->insy_bufhead = 0;
				}
			}
		}
		if ( (insy_twd->insy_currlen >= INSYTRACK_MSG_MARK_LEN
				+ INSYTRACK_MSG_TYPE_LEN)
			&& (insy_msg_type == 0xFFFFFFFF))
		{
			insy_msg_type = libinsytrack_extract_number (
				&(insy_twd->insy_buf[insy_twd->insy_bufhead
					+ INSYTRACK_MSG_MARK_LEN]),
				INSYTRACK_MSG_TYPE_LEN);
			if ( (insy_msg_type != INSYTRACK_MESSAGE_DATA)
				&& (insy_msg_type != INSYTRACK_MESSAGE_LAST) )
			{
				break;
			}
		}
		if ( (insy_twd->insy_currlen >= INSYTRACK_MSG_MARK_LEN
				+ INSYTRACK_MSG_TYPE_LEN
				+ INSYTRACK_MSGID_LENGTH_LEN)
			&& (insy_msgid_length == 0xFFFFFFFF))
		{
			insy_msgid_length = libinsytrack_extract_number (
				&(insy_twd->insy_buf[insy_twd->insy_bufhead
					+ INSYTRACK_MSG_MARK_LEN
					+ INSYTRACK_MSG_TYPE_LEN]),
				INSYTRACK_MSGID_LENGTH_LEN);
		}
		if ( (insy_twd->insy_currlen >= INSYTRACK_MSG_MARK_LEN
				+ INSYTRACK_MSG_TYPE_LEN
				+ INSYTRACK_MSGID_LENGTH_LEN
				+ INSYTRACK_MSGBODY_LENGTH_LEN)
			&& (insy_msg_length == 0xFFFFFFFF))
		{
			insy_msg_length = libinsytrack_extract_number (
				&(insy_twd->insy_buf[insy_twd->insy_bufhead
					+ INSYTRACK_MSG_MARK_LEN
					+ INSYTRACK_MSG_TYPE_LEN
					+ INSYTRACK_MSGID_LENGTH_LEN]),
				INSYTRACK_MSGBODY_LENGTH_LEN);
		}
		/* a marker so we don't leave if we didn't call reading at all: */
		insy_bytes_read = 0;
		/* read new bytes only when there is space and we don't
		have enough bytes for at least one message */
		if ( (!((insy_twd->insy_currlen >= INSYTRACK_MSG_MARK_LEN
			+ INSYTRACK_MSG_TYPE_LEN
			+ INSYTRACK_MSGID_LENGTH_LEN
			+ INSYTRACK_MSGBODY_LENGTH_LEN
			+ insy_msgid_length
			+ insy_msg_length)
			&& (insy_msgid_length != 0xFFFFFFFF)
			&& (insy_msg_length != 0xFFFFFFFF)))
			&& (insy_twd->insy_bufsize > insy_twd->insy_bufhead
			+ insy_twd->insy_currlen) )
		{
#ifdef INSYTRACK_HAVE_SELECT
			/* select() can destroy its parameters, so reset every time */
			FD_ZERO (&insy_set_r);
			FD_SET (insy_twd->insy_client_socket, &insy_set_r);	/* warnings are inside this macro */
			FD_ZERO (&insy_set_ex);
			FD_SET (insy_twd->insy_client_socket, &insy_set_ex);	/* warnings are inside this macro */
			insy_tv.tv_sec = INSYTRACK_MAX_SELECT_SECONDS;
			insy_tv.tv_usec = 0;

			if ( select (insy_twd->insy_client_socket+1, &insy_set_r,
				NULL, &insy_set_ex, &insy_tv) > 0 )
			{
				if ( FD_ISSET (insy_twd->insy_client_socket, &insy_set_ex) )
				{
					break;
				}
#endif
				insy_bytes_read = recvfrom (insy_twd->insy_client_socket,
					&(insy_twd->insy_buf[insy_twd->insy_bufhead
						+ insy_twd->insy_currlen]),
					insy_twd->insy_bufsize - (insy_twd->insy_bufhead
						+ insy_twd->insy_currlen),
					0, NULL, NULL);
				if ( insy_bytes_read <= 0 )
				{
					break;
				}
				insy_twd->insy_currlen += (size_t)insy_bytes_read;
#ifdef INSYTRACK_HAVE_SELECT
			}
			else
			{
				if ( insy_twd->insy_server->insy_shutting_down != 0 )
				{
					break;
				}
			}
#endif
		}
		if ( (insy_twd->insy_currlen >= INSYTRACK_MSG_MARK_LEN
			+ INSYTRACK_MSG_TYPE_LEN
			+ INSYTRACK_MSGID_LENGTH_LEN
			+ INSYTRACK_MSGBODY_LENGTH_LEN
			+ insy_msgid_length
			+ insy_msg_length)
			&& (insy_msgid_length != 0xFFFFFFFF)
			&& (insy_msg_length != 0xFFFFFFFF) )
		{
			insy_total_length = INSYTRACK_MSG_MARK_LEN
				+ INSYTRACK_MSG_TYPE_LEN
				+ INSYTRACK_MSGID_LENGTH_LEN
				+ INSYTRACK_MSGBODY_LENGTH_LEN
				+ insy_msgid_length
				+ insy_msg_length;
			if ( insy_msg != NULL )
			{
				insy_msg->insy_type.insy_type_int
					= insy_msg_type;
				insy_msg->insy_msgid_len.insy_msgid_len_int
					= insy_msgid_length;
				insy_msg->insy_msgbody_len.insy_msgbody_len_int
					= insy_msg_length;
				if ( (insy_msg->insy_msgid != NULL)
					&& (insy_msgid_length > 0) )
				{
#ifdef HAVE_MEMCPY
					memcpy (insy_msg->insy_msgid,
						&(insy_twd->insy_buf[insy_twd->insy_bufhead
							+ INSYTRACK_MSG_MARK_LEN
							+ INSYTRACK_MSG_TYPE_LEN
							+ INSYTRACK_MSGID_LENGTH_LEN
							+ INSYTRACK_MSGBODY_LENGTH_LEN]),
						insy_msgid_length);
#else
					for ( insy_offset = 0;
						insy_offset < insy_msgid_length;
						insy_offset++ )
					{
						((unsigned char *)insy_msg->insy_msgid)[insy_offset]
							= insy_twd->insy_buf[
								insy_twd->insy_bufhead
								+ INSYTRACK_MSG_MARK_LEN
								+ INSYTRACK_MSG_TYPE_LEN
								+ INSYTRACK_MSGID_LENGTH_LEN
								+ INSYTRACK_MSGBODY_LENGTH_LEN
								+ insy_offset];
					}
#endif
				}
				if ( (insy_msg->insy_msgbody != NULL) && (insy_msg_length > 0) )
				{
#ifdef HAVE_MEMCPY
					memcpy (insy_msg->insy_msgbody,
						&(insy_twd->insy_buf[insy_twd->insy_bufhead
							+ INSYTRACK_MSG_MARK_LEN
							+ INSYTRACK_MSG_TYPE_LEN
							+ INSYTRACK_MSGID_LENGTH_LEN
							+ INSYTRACK_MSGBODY_LENGTH_LEN
							+ insy_msgid_length]),
						insy_msg_length);
#else
					for ( insy_offset = 0;
						insy_offset < insy_msg_length;
						insy_offset++ )
					{
						((unsigned char *)insy_msg->insy_msgbody)[insy_offset]
							= insy_twd->insy_buf[
								insy_twd->insy_bufhead
								+ INSYTRACK_MSG_MARK_LEN
								+ INSYTRACK_MSG_TYPE_LEN
								+ INSYTRACK_MSGID_LENGTH_LEN
								+ INSYTRACK_MSGBODY_LENGTH_LEN
								+ insy_msgid_length
								+ insy_offset];
					}
#endif
				}
			}
#ifdef HAVE_MEMMOVE
			memmove (insy_twd->insy_buf,
				&(insy_twd->insy_buf[insy_total_length]),
				insy_twd->insy_currlen - insy_total_length);
#else
			for ( insy_offset = 0;
				insy_offset < insy_twd->insy_currlen - insy_total_length;
				insy_offset++ )
			{
				insy_twd->insy_buf[insy_offset]
					= insy_twd->insy_buf[insy_total_length + insy_offset];
			}
#endif
			insy_twd->insy_bufhead = 0;
			insy_twd->insy_currlen -= insy_total_length;
			insy_res = INSYTRACK_MSGRET_SUCCESS;
			break;
		}
		else if ( insy_bytes_read < 0 )
		{
			break;
		}
	}
	while ( insy_twd->insy_server->insy_shutting_down == 0 );
	return insy_res;
}

/* ======================================================================== */

/**
 * The connection worker thread.
 * @param insy_param The parameter passed by pthread_create().
 * @return NULL.
 */
void * insytrack_worker (
#ifdef INSYTRACK_ANSIC
	void * insy_param)
#else
	insy_param)
	void * insy_param;
#endif
{
	struct insytrack_worker_data * insy_thread_data;
	insytrack_message_t insy_msg;
	insytrack_msgret_t insy_msgret;
	insytrack_callback_return_t insy_callret;

	insy_thread_data = (struct insytrack_worker_data *) insy_param;

	if ( insy_thread_data == NULL )
	{
		/* don't deatch, wait for join */
		/*pthread_detach (pthread_self ());*/
		return NULL;
	}

	if ( insy_thread_data->insy_server == NULL )
	{
		/* don't deatch, wait for join */
		/*pthread_detach (pthread_self ());*/
		return NULL;
	}

	insy_thread_data->insy_is_running = 1;

	/* wait for full initialization (simply lock & unlock): */
	pthread_mutex_lock (&(insy_thread_data->insy_server->insy_list_mutex));
	pthread_mutex_unlock (&(insy_thread_data->insy_server->insy_list_mutex));

	/* serve the client's requests */
	if ( insy_thread_data->insy_client_socket >= 0 )
	{
		insy_thread_data->insy_buf = (unsigned char *) malloc (INSYTRACK_MSG_BUF_SIZE);
		insy_thread_data->insy_bufhead = 0;
		insy_thread_data->insy_bufsize = INSYTRACK_MSG_BUF_SIZE;
		insy_thread_data->insy_currlen = 0;
		insy_msg.insy_msgid = malloc (INSYTRACK_MSGID_MAXLEN+1);
		insy_msg.insy_msgbody = malloc (INSYTRACK_MSGBODY_MAXLEN+1);
		if ( (insy_thread_data->insy_buf != NULL)
			&& (insy_msg.insy_msgid != NULL)
			&& (insy_msg.insy_msgbody != NULL) )
		{
			do
			{
				insy_msg.insy_type.insy_type_int = 0xFFFFFFFF;
				insy_msg.insy_msgid_len.insy_msgid_len_int = 0;
				insy_msg.insy_msgbody_len.insy_msgbody_len_int = 0;
				((char *)insy_msg.insy_msgid)[0] = '\0';
				((char *)insy_msg.insy_msgbody)[0] = '\0';

				insy_msgret = libinsytrack_server_read_message (
					insy_thread_data, &insy_msg);
				if ( insy_msgret == INSYTRACK_MSGRET_SUCCESS )
				{
					if ( insy_thread_data->insy_server->insy_msgrecvd_callback != NULL )
					{
						insy_callret = (*(insy_thread_data->insy_server->insy_msgrecvd_callback))
							(&insy_msg, insy_thread_data->insy_server->insy_msgrecvd_callback_data);
						if ( insy_callret == INSYTRACK_CALLBACK_ABORT )
						{
							break;
						}
					}
					/* If no callback is set, just throw away the message.
					Useful for simply checking the communication. */
				}
				else
				{
					break;
				}
			} while (insy_thread_data->insy_server->insy_shutting_down == 0);
		}

		if ( insy_msg.insy_msgbody != NULL )
		{
			free (insy_msg.insy_msgbody);
		}
		if ( insy_msg.insy_msgid != NULL )
		{
			free (insy_msg.insy_msgid);
		}
		if ( insy_thread_data->insy_buf != NULL )
		{
			free (insy_thread_data->insy_buf);
		}

		if ( insy_thread_data->insy_server->insy_disconn_callback != NULL )
		{
			(*(insy_thread_data->insy_server->insy_disconn_callback)) (
				insy_thread_data->insy_client_addr,
				insy_thread_data->insy_client_addr_len,
				insy_thread_data->insy_server->insy_disconn_callback_data);
		}
		if ( insy_thread_data->insy_client_addr != NULL )
		{
			free (insy_thread_data->insy_client_addr);
		}
#ifdef HAVE_CLOSE
		close (insy_thread_data->insy_client_socket);
#endif
	}
	insy_thread_data->insy_is_running = 0;
	insytrack_remove_worker (insy_thread_data->insy_server, pthread_self ());

	return NULL;
}

