/*
 * InSyTrack - software for tracking program flow, server files.
 *	-- server threads' list.
 *
 * Copyright (C) 2012-2013 Bogdan Drozdowski, bogdandr (at) op.pl
 * License: GNU Lesser General Public License, v3+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foudation:
 *		Free Software Foundation
 *		51 Franklin Street, Fifth Floor
 *		Boston, MA 02110-1301
 *		USA
 */

#include "insytrack_cfg.h"

#include "insytrack_server.h"
#include "libinsytrack_server_init.h"
#include "libinsytrack_server_thread.h"
#include "libinsytrack_server_tlist.h"

#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif

#ifdef HAVE_STRING_H
# if ((!defined STDC_HEADERS) || (!STDC_HEADERS)) && (defined HAVE_MEMORY_H)
#  include <memory.h>
# endif
# include <string.h>	/* memcpy() */
#endif

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_MALLOC_H
# include <malloc.h>
#endif

/* ======================================================================== */

/**
 * Creates a new worker thread and adds it to the server's list.
 * @param insy_server The server to add the thread to.
 * @param insy_twd The thread's working data.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t insytrack_create_worker (
#ifdef INSYTRACK_ANSIC
	struct insytrack_server * insy_server,
	struct insytrack_worker_data * insy_twd)
#else
	insy_server, insy_twd)
	struct insytrack_server * insy_server;
	struct insytrack_worker_data * insy_twd;
#endif
{
	int insy_result;
	pthread_t insy_new_thread;
	struct insytrack_worker_data * insy_priv_twd;
	struct insytrack_thread_list * insy_list_elem;
	struct insytrack_thread_list * insy_curr_elem;
#ifndef HAVE_MEMCPY
	socklen_t  i;
#endif

	if ( (insy_server == NULL) || (insy_twd == NULL) )
	{
		return INSYTRACKS_PARAM_ERROR;
	}

	/* allocate a new struct insytrack_worker_data to copy the data */
	insy_priv_twd = (struct insytrack_worker_data *) malloc (
		sizeof (struct insytrack_worker_data));
	if ( insy_priv_twd == NULL )
	{
		return INSYTRACKS_MALLOC_ERROR;
	}

	/* create a new element for the list of threads */
	insy_list_elem = (struct insytrack_thread_list *) malloc (
		sizeof (struct insytrack_thread_list));
	if ( insy_list_elem == NULL )
	{
		free (insy_priv_twd);
		return INSYTRACKS_MALLOC_ERROR;
	}

	/* lock here so the new thread doesn't start immediately */
	pthread_mutex_lock (&(insy_server->insy_list_mutex));
	insy_priv_twd->insy_client_socket = insy_twd->insy_client_socket;
	insy_priv_twd->insy_server = insy_twd->insy_server;
	if ( insy_twd->insy_client_addr_len > 0 )
	{
		insy_priv_twd->insy_client_addr = (struct sockaddr *) malloc (
			insy_twd->insy_client_addr_len);
		if ( insy_priv_twd->insy_client_addr != NULL )
		{
#ifdef HAVE_MEMCPY
			memcpy (insy_priv_twd->insy_client_addr,
				insy_twd->insy_client_addr,
				insy_twd->insy_client_addr_len);
#else
			for ( i = 0; i < insy_twd->insy_client_addr_len; i++ )
			{
				((char*)insy_priv_twd->insy_client_addr)[i]
					= ((char*)insy_twd->insy_client_addr)[i];
			}
#endif
		}
	}
	else
	{
		insy_priv_twd->insy_client_addr = NULL;
	}
	insy_priv_twd->insy_client_addr_len = insy_twd->insy_client_addr_len;
	insy_priv_twd->insy_is_running = 0;

	/* create the thread */
	insy_result = pthread_create (&insy_new_thread, NULL,
		&insytrack_worker, insy_priv_twd);
	if ( insy_result != 0 )
	{
		free (insy_priv_twd);
		free (insy_list_elem);
		return INSYTRACKS_THREAD_ERROR;
	}
	insy_list_elem->insy_thread_id = insy_new_thread;
	insy_list_elem->insy_thread_data = insy_priv_twd;
	insy_list_elem->insy_next = NULL;

	/* add the new thread to the list */
	insy_curr_elem = insy_server->insy_tlist_head;
	if ( insy_curr_elem != NULL )
	{
		while (insy_curr_elem->insy_next != NULL)
		{
			insy_curr_elem = insy_curr_elem->insy_next;
		}
		/* here wre're at the end of the list */
		insy_curr_elem->insy_next = insy_list_elem;
		insy_list_elem->insy_prev = insy_curr_elem;
	}
	else
	{
		/* the list is empty - add the first element */
		insy_list_elem->insy_prev = NULL;
		insy_server->insy_tlist_head = insy_list_elem;
	}
	pthread_mutex_unlock (&(insy_server->insy_list_mutex));

	return INSYTRACKS_OK;
}

/* ======================================================================== */

/**
 * Removes a worker thread from the server's list.
 * @param insy_server The server to remove the thread from.
 * @param insy_thread The thread to remove.
 * @return an insytrack_server_error_t value specifying the
 *	status of the operation.
 */
insytrack_server_error_t insytrack_remove_worker (
#ifdef INSYTRACK_ANSIC
	struct insytrack_server * insy_server, pthread_t insy_thread)
#else
	insy_server, insy_thread)
	struct insytrack_server * insy_server;
	pthread_t insy_thread;
#endif
{
	struct insytrack_thread_list * insy_curr_elem;

	if ( insy_server == NULL )
	{
		return INSYTRACKS_PARAM_ERROR;
	}

	pthread_mutex_lock (&(insy_server->insy_list_mutex));
	insy_curr_elem = insy_server->insy_tlist_head;
	while ( insy_curr_elem != NULL )
	{
		if ( pthread_equal (insy_curr_elem->insy_thread_id,
			insy_thread) != 0 )
		{
			if ( insy_curr_elem->insy_prev != NULL )
			{
				insy_curr_elem->insy_prev->insy_next
					= insy_curr_elem->insy_next;
			}
			if ( insy_curr_elem->insy_next != NULL )
			{
				insy_curr_elem->insy_next->insy_prev
					= insy_curr_elem->insy_prev;
			}
			if ( insy_curr_elem->insy_thread_data != NULL )
			{
				free (insy_curr_elem->insy_thread_data);
			}
			if ( (insy_curr_elem == insy_server->insy_tlist_head)
				&& (insy_curr_elem->insy_prev == NULL)
				&& (insy_curr_elem->insy_next == NULL)
				)
			{
				/* mark the list as empty so the main thread
				can exit */
				insy_server->insy_tlist_head = NULL;
			}
			free (insy_curr_elem);
			/* Deatch the thread, so it will be cleaned up.
			   This allows the thread to remove itself from the
			   list without being joined. */
			pthread_detach (insy_thread);
			break;
		}
		insy_curr_elem = insy_curr_elem->insy_next;
	}
	pthread_mutex_unlock (&(insy_server->insy_list_mutex));

	return INSYTRACKS_OK;
}
